# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Geoips module for setting up logging handlers with a specified verbosity."""

import logging
import sys
from textwrap import wrap
from os import getenv
from socket import gethostname

from geoips.filenames.base_paths import PATHS as gpaths


logging.captureWarnings(True)


def log_with_emphasis(print_func, *messages):
    """Print messages boxed in asterisks using the specified print function.

    Print one or more messages using the specified print function. The messages
    will be surrounded in asterisks. Long messages will be word wrapped to fit
    within a maximum width of 74 characters, save for any individual words that
    are over 74 characters which will not be broken.

    Parameters
    ----------
    print_func: func
        An instance of a function that prints (e.g. ``logging.debug``, ``logging.info``,
        etc, or the ``print`` function itself).
    messages: one or more strings
        The messages to be logged with emphasis
    """
    wrapped_messages = []
    # Even though only strings should be passed here, we are going to allow any type of
    # object to be passed for the time being. Best not to overthink the situation and
    # just cast for the time being.
    messages = filter(lambda s: len(str(s)) > 0, messages)

    for message in messages:
        # wrap the message to a specified length
        wrapped_messages += wrap(
            message, width=74, break_long_words=False, break_on_hyphens=False
        )
    try:
        max_message_len = min(74, max([len(wmessage) for wmessage in wrapped_messages]))
    except ValueError as e:
        raise ValueError(
            "No proper messages were provided for logging.\n"
            + "Make sure the messages are not all empty strings."
        ) from e
    # adding +6 to max_message_len as we add '** ' and ' **' pre/post-fixes (6 chars)
    print_func("*" * (max_message_len + 6))
    for wrapped_message in wrapped_messages:
        # for each of the wrapped messages, if the length of such message is less
        # than max message length, add some whitespace to make some things match,
        # this is what the 'ljust(len)' function does
        print_func(f"** {wrapped_message.ljust(max_message_len)} **")
    print_func("*" * (max_message_len + 6))
    print_func("\n")


class LogLevelAdder:
    """Create a callable that can add a new logging level.

    Initialize simply as `add_log_level = LogLevelAdder()` and add a new level by
    calling `add_log_level` with the name of the new log level and its level number.

    For further call information, see the docstring for LogLevelAdder.__call__().
    """

    def __call__(self, level_name, level_num):
        """Comprehensively adds a new logging level to the `logging` module.

        Comprehensively adds a new logging level named `level_name` to the `logging`
        module and the currently configured logging class. The new logging level's
        precidence is set by `level_num`. For example, a level_num of `15` would create
        a logging level that falls between `logging.DEBUG` and `logging.INFO` whose
        level_num are `10` and `20`, respectively.

        Calling this function will add the following to the logging class:
        - A new log level named `logging.{level_name}`
        - A new logging function called `logging.{level_name.lower()]}`

        To avoid conflicts between the log level attribute and the logging function,
        `level_name` must not be entirely lowercase.

        Parameters
        ----------
        level_name : str
            Name for the new log level. This must not be completely lowercase.
        level_num : int
            Numeric precidence of the new log level.

        `level_name` becomes an attribute of the `logging` module with the value
        `level_num`. `methodName` becomes a convenience method for both `logging`
        itself and the class returned by `logging.getLoggerClass()` (usually just
        `logging.Logger`). If `methodName` is not specified, `level_name.lower()` is
        used.

        To avoid accidental clobberings of existing attributes, this method will
        raise an `AttributeError` if the level name is already an attribute of the
        `logging` module or if the method name is already present

        Example
        -------
        >>> add_logging_level = LogLevelAdder()
        >>> add_logging_level('TRACE', logging.DEBUG - 5)
        >>> logging.getLogger(__name__).setLevel("TRACE")
        >>> logging.getLogger(__name__).trace('that worked')
        >>> logging.trace('so did this')
        >>> logging.TRACE
        5

        """
        methodName = level_name.lower()

        # Add the log level
        if hasattr(logging, level_name):
            if getattr(logging, level_name) != level_num:
                raise AttributeError(
                    f"{level_name} level already defined in logging module and its "
                    f"value ({getattr(logging, level_name)}) differs from the "
                    f"requested value ({level_num})."
                )
        else:
            logging.addLevelName(level_num, level_name)
            setattr(logging, level_name, level_num)

        # Add the logging method to the current root logger
        logToRoot = self._get_logToRoot(level_num)
        if hasattr(logging, methodName):
            if getattr(logging, methodName) != logToRoot:
                raise AttributeError(
                    f"{methodName} method already defined in logging module and "
                    f"differs from the requested method."
                )
        else:
            setattr(logging, methodName, logToRoot)

        # Add the logging method to the logger class so it gets attached to new logger
        # instances moving forward
        loggerClass = logging.getLoggerClass()
        logForLevel = self._get_logForLevel(level_num)
        if hasattr(loggerClass, methodName):
            if getattr(loggerClass, methodName) != logForLevel:
                raise AttributeError(
                    f"{methodName} method already defined in logger class and differs "
                    f"from the requested method."
                )
        else:
            setattr(loggerClass, methodName, logForLevel)

    def _get_logForLevel(self, level_num):
        if not hasattr(self, "_logForLevel_funcs"):
            self._logForLevel_funcs = {}

        if level_num not in self._logForLevel_funcs:

            def logForLevel(self, message, *args, **kwargs):
                if self.isEnabledFor(level_num):
                    self._log(level_num, message, args, **kwargs)

            self._logForLevel_funcs[level_num] = logForLevel
        return self._logForLevel_funcs[level_num]

    def _get_logToRoot(self, level_num):
        if not hasattr(self, "_logToRoot_funcs"):
            self._logToRoot_funcs = {}

        if level_num not in self._logToRoot_funcs:

            def logToRoot(message, *args, **kwargs):
                logging.log(level_num, message, *args, **kwargs)

            self._logToRoot_funcs[level_num] = logToRoot
        return self._logToRoot_funcs[level_num]


add_logging_level = LogLevelAdder()


def setup_logging(logging_level=None, fmt_string=None, datefmt_string=None):
    """Get a new logger instance for GeoIPS.

    Get a new logger instance for GeoIPS. This will set the logger's logging level, its
    formatter, and add a `StreamHandler` pointing to `sys.stdout`.

    This is most often used at the top-level of an applcation to set up the root logger
    for the application (e.g. in a command line script). Once configured, the root
    logger's properties will be inherited by lower-level logger instances. So, to use
    the same logging configuration in submodules, simply instantiate a new logger
    instance via `LOG = logging.getLogger(__name__)` and it will behave the same as the
    root logger.

    Parameters
    ----------
    logging_level : str, default=None
        Sets the minimum log level for which log output will be written to stdout.
        If None, will default to value set in base_paths.  This allows using
        GEOIPS_LOGGING_LEVEL to override default.
    fmt_string: str, default=None
        Explicitly specify the log formatter format string. If None, default
        to the value set in geoips/filenames/base_paths (which in the future
        will be replaced with a GeoIPS config file).
    datefmt_string: str, default=None
        Explicitly specify the log formatter datefmt string. If None, default
        to the value set in geoips/filenames/base_paths (which in the future
        will be replaced with a GeoIPS config file).
    """
    log = logging.getLogger()

    ###########################################################################
    # Identify the current requested logging level

    # If logging_level was not specified, default to base_path specification here.
    if not logging_level:
        logging_level = gpaths["GEOIPS_LOGGING_LEVEL"]
    log.setLevel(getattr(logging, logging_level.upper()))

    ###########################################################################
    # Identify the current requested format and date format strings for
    # log formatter.

    if fmt_string is None:
        # Note that any format specifications supported directly by the
        # logging.Formatter fmt keyword argument are specified as %(fmtspec)
        # Username and hostname were not directly supported by logging.Formatter,
        # so we added the capability to include them via os.getenv and
        # socket.gethostname and f-string formatting.
        fmt_string = gpaths["GEOIPS_LOGGING_FMT_STRING"]
    if datefmt_string is None:
        datefmt_string = gpaths["GEOIPS_LOGGING_DATEFMT_STRING"]

    ###########################################################################
    # Now replace any variables specified as {varname} within the format string.

    username = getenv("USER")
    hostname = gethostname()

    # These are the variables explicitly supported in the format string, no other
    # variables specified with {varname} are supported.
    fmt_vars = {"username": username, "hostname": hostname}
    # Identify which of the above variables are actually in the format string.
    fmt_kwargs = {x: y for x, y in fmt_vars.items() if x in fmt_string}
    # Pass each of the variables specified in the format string as keyword
    # arguments to the format method of the string.  Ie,
    # fmt_string.format(username=username, hostname=hostname)
    fmt_string = fmt_string.format(**fmt_kwargs)

    ###########################################################################
    # Set the actual log formatter using fmt_string and datefmt_string.

    fmt = logging.Formatter(fmt=fmt_string, datefmt=datefmt_string)

    ###########################################################################
    # Set the log handlers if they are not already set

    if not log.handlers:
        stream_hndlr = logging.StreamHandler(sys.stdout)
        stream_hndlr.setFormatter(fmt)
        stream_hndlr.setLevel(logging.INFO)
        log.addHandler(stream_hndlr)

    ###########################################################################
    # Print username and host at the info level at the beginning of the log output

    log.info(f"USER: {username}")
    log.info(f"HOST: {hostname}")

    ###########################################################################
    # Return the resulting log for use throughout GeoIPS

    return log



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Unit tests for sector plugins."""

import pytest
from geoips import interfaces
from geoips.interfaces.yaml_based.sectors import SectorPluginBase


def test_get_all_sectors():
    """Attempt to load all available sectors."""
    interfaces.sectors.get_plugins()


def gen_ids(sector_dict):
    """Generate ids from the name key in each sector."""
    return sector_dict["name"]


@pytest.mark.parametrize("sector", interfaces.sectors.get_plugins(), ids=gen_ids)
def test_sector(sector):
    """Ensure that a sector has the correct attributes."""
    assert isinstance(sector, SectorPluginBase)
    assert sector.interface == "sectors"
    assert sector.name
    assert sector.family
    assert sector.__doc__
    assert sector.docstring == sector.__doc__
    # Should add this back after completing #301
    # assert isinstance(sector.area_definition, pyresample.AreaDefinition)



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Classes for geometry operations allowing masked data in swath corners.

This package uses the pyresample geometry package as base classes:

# pyresample, Resampling of remote sensing image data in python
#
# Copyright (C) 2010-2015
#
# Authors:
#    Esben S. Nielsen
#    Thomas Lavergne
"""
from __future__ import absolute_import

# Python Standard Libraries
import warnings
import math
import logging

# Installed Libraries
import numpy as np
from pyproj import Geod
from pyresample import utils
from pyresample.geometry import SwathDefinition, CoordinateDefinition

log = logging.getLogger(__name__)

interface = None

EPSILON = 0.0000001


class MaskedCornersSwathDefinition(SwathDefinition):
    """Swath defined by lons and lats.

    Allows datasets with potentially masked data in the corners.

    Parameters
    ----------
    lons : numpy array
        Longitude values
    lats : numpy array
        Latitude values
    nprocs : int, optional
        Number of processor cores to be used for calculations.

    Attributes
    ----------
    shape : tuple
        Swath shape
    size : int
        Number of elements in swath
    ndims : int
        Swath dimensions

    Properties
    ----------
    lons : object
        Swath lons
    lats : object
        Swath lats
    cartesian_coords : object
        Swath cartesian coordinates
    """

    def __init__(self, lons, lats, nprocs=1):
        """Initialize class."""
        if lons.shape != lats.shape:
            raise ValueError("lon and lat arrays must have same shape")
        elif lons.ndim > 2:
            raise ValueError("Only 1 and 2 dimensional swaths are allowed")

        if lons.shape == lats.shape and lons.dtype == lats.dtype:
            self.shape = lons.shape
            self.size = lons.size
            self.ndim = lons.ndim
            self.dtype = lons.dtype
        else:
            raise ValueError(
                (
                    "%s must be created with either "
                    "lon/lats of the same shape with same dtype"
                )
                % self.__class__.__name__
            )

        lats_type = type(lats)
        lons_type = type(lons)

        if lats_type != lons_type:
            raise TypeError("lons and lats must be of same type")
        elif lons is not None:
            if lons.shape != lats.shape:
                raise ValueError("lons and lats must have same shape")

        self.nprocs = nprocs

        # check the latitutes
        if lats is not None and ((lats.min() < -90.0 or lats.max() > +90.0)):
            # throw exception
            raise ValueError("Some latitudes are outside the [-90.;+90] validity range")
        else:
            self.lats = lats

        # check the longitudes
        if lons is not None and ((lons.min() < -180.0 or lons.max() >= +180.0)):
            # issue warning
            warnings.warn(
                "All geometry objects expect longitudes in the [-180:+180[ range. "
                "We will now automatically wrap your longitudes into [-180:+180[, "
                "and continue. "
                "To avoid this warning next time, use routine utils.wrap_longitudes()."
            )
            # wrap longitudes to [-180;+180[
            self.lons = utils.wrap_longitudes(lons)
        else:
            self.lons = lons

        self.cartesian_coords = None

    def intersection(self, other):
        """Return current area intersection polygon corners.

        *other* allows for potentially masked data in the corners.

        Parameters
        ----------
        other : object
            Instance of subclass of BaseDefinition

        Returns
        -------
        (corner1, corner2, corner3, corner4) : tuple of points
        """
        from pyresample.spherical_geometry import intersection_polygon

        # This was failing if all the corners of the
        #       area_definition fell inside the data box definition.
        #       watch for false positives
        # This DOES NOT WORK for over the pole...
        retcorners = intersection_polygon(self.corners, other.corners)
        allselfcornersin = False
        allothercornersin = False
        if not retcorners:
            # Only try these if intersection_polygon didn't return anything.
            for i in self.corners:
                if planar_point_inside(i, other.corners):
                    allselfcornersin = True
                else:
                    allselfcornersin = False
            for i in other.corners:
                if planar_point_inside(i, self.corners):
                    allothercornersin = True
                else:
                    allothercornersin = False

            if allselfcornersin:
                return self.corners
            if allothercornersin:
                return other.corners
        return retcorners

    def overlaps_minmaxlatlon(self, other):
        """Test current area overlaps *other* area.

        This is based solely on the min/max lat/lon of areas, assuming the
         boundaries to be along lat/lon lines.

        Parameters
        ----------
        other : object
            Instance of subclass of BaseDefinition

        Returns
        -------
        overlaps : bool
        """
        self_corners = get_2d_false_corners(self)
        other_corners = get_2d_false_corners(other)
        log.info("    Swath 2d False Corners: " + str(self_corners))
        log.info("    Other 2d False Corners: " + str(other_corners))

        for i in self_corners:
            if planar_point_inside(i, other_corners):
                return True
        for i in other_corners:
            if planar_point_inside(i, self_corners):
                return True
        return False

    @property
    def corners(self):
        """Return current area corners."""
        try:
            # Try to just set normal CoordinateDefinition corners
            #    (Which doesn't work with bad vals in corners)
            return super(CoordinateDefinition, self).corners
        except ValueError:
            # print '        Corners failed on CoordinateDefinition, try falsecorners'
            pass

        lons, lats = self.get_lonlats()

        # Determine which rows and columns contain good data
        rows = lons.any(axis=1)
        cols = lons.any(axis=0)

        # Get the minimum and maximum row and column that contain good data
        good_row_inds = np.where(~rows.mask)[0]
        min_row = good_row_inds.min()
        max_row = good_row_inds.max()

        good_col_inds = np.where(~cols.mask)[0]
        min_col = good_col_inds.min()
        max_col = good_col_inds.max()

        log.info(
            "    USING FALSE CORNERS!! setting corners. min row/col: "
            + str(min_row)
            + " "
            + str(min_col)
            + " "
            + "max row/col: "
            + str(max_row)
            + " "
            + str(max_col)
            + " "
            + "shape: "
            + str(lons.shape)
        )

        # from .spherical import SCoordinate as Coordinate
        # from .spherical import Arc
        from pyresample.spherical_geometry import Coordinate, Arc

        # Calculate the eight possible corners and produce arcs for each pair
        # Corners for top side
        # Right side was failing with Divide by Zero error for NCC data because there
        # was a single good point in the max_col.  Keep incrementing or decrementing
        # until good.min doesn't equal good.max
        good = np.where(~lons[min_row, :].mask)[0]
        tries = 0
        while tries < 20 and good.min() == good.max():
            # print 'good.min() can\'t equal good.max() for top side, incrementing
            # min_row! Would have failed with ZeroDivisionError before!'
            min_row += 1
            tries += 1
            good = np.where(~lons[min_row, :].mask)[0]
        top_corners = [
            Coordinate(*self.get_lonlat(min_row, good.min())),
            Coordinate(*self.get_lonlat(min_row, good.max())),
        ]
        top_arc = Arc(top_corners[0], top_corners[1])

        # Corners for bottom side
        good = np.where(~lons[max_row, :].mask)[0]
        tries = 0
        while tries < 20 and good.min() == good.max():
            # print 'good.min() can\'t equal good.max() for bottom side, decrementing
            # max_row! Would have failed with ZeroDivisionError before!'
            max_row -= 1
            tries += 1
            good = np.where(~lons[max_row, :].mask)[0]
        bot_corners = [
            Coordinate(*self.get_lonlat(max_row, good.min())),
            Coordinate(*self.get_lonlat(max_row, good.max())),
        ]
        bot_arc = Arc(bot_corners[0], bot_corners[1])

        # Corners for left side
        good = np.where(~lons[:, min_col].mask)[0]
        tries = 0
        while tries < 20 and good.min() == good.max():
            # print 'good.min() can\'t equal good.max() for left side, incrementing
            # min_col! Would have failed with ZeroDivisionError before!'
            min_col += 1
            tries += 1
            good = np.where(~lons[:, min_col].mask)[0]
        left_corners = [
            Coordinate(*self.get_lonlat(good.min(), min_col)),
            Coordinate(*self.get_lonlat(good.max(), min_col)),
        ]
        left_arc = Arc(left_corners[0], left_corners[1])

        # Corners for right side
        good = np.where(~lons[:, max_col].mask)[0]
        tries = 0
        while tries < 20 and good.min() == good.max():
            # print 'good.min() can\'t equal good.max() for right side, decrementing
            # max_col! Would have failed with ZeroDivisionError before!'
            max_col -= 1
            tries += 1
            good = np.where(~lons[:, max_col].mask)[0]
        right_corners = [
            Coordinate(*self.get_lonlat(good.min(), max_col)),
            Coordinate(*self.get_lonlat(good.max(), max_col)),
        ]
        right_arc = Arc(right_corners[0], right_corners[1])

        # Calculate the four false corners
        _corners = []
        # Top left false corner
        top_intersections = top_arc.intersections(left_arc)
        dists = [inter.distance(top_corners[0]) for inter in top_intersections]
        if dists[0] < dists[1]:
            _corners.append(top_intersections[0])
        else:
            _corners.append(top_intersections[1])
        # Top right false corner
        top_intersections = top_arc.intersections(right_arc)
        dists = [inter.distance(top_corners[1]) for inter in top_intersections]
        if dists[0] < dists[1]:
            _corners.append(top_intersections[0])
        else:
            _corners.append(top_intersections[1])
        # Bottom right false corner
        bot_intersections = bot_arc.intersections(right_arc)
        dists = [inter.distance(bot_corners[1]) for inter in bot_intersections]
        if dists[0] < dists[1]:
            _corners.append(bot_intersections[0])
        else:
            _corners.append(bot_intersections[1])
        # Bottom left false corner
        bot_intersections = bot_arc.intersections(left_arc)
        dists = [inter.distance(bot_corners[0]) for inter in bot_intersections]
        if dists[0] < dists[1]:
            _corners.append(bot_intersections[0])
        else:
            _corners.append(bot_intersections[1])
        return _corners

    def get_bounding_box_lonlats(self, npts=100):
        """Return lon/lats along bounding Arcs.

        Parameters
        ----------
        npts: int
            Number of points to return along each line

        Returns
        -------
        (top, right, bottom, left) : 4 tuples containing lists
                                    of len npts of lons/lats
        retval = (list(tplons),list(tplats)),
                 (list(rtlons),list(rtlats)),
                 (list(btlons),list(btlats)),
                 (list(ltlons),list(ltlats))

        eg for n=3
           ([tplon0,tplon1,tplon2],[tplat0,tplat1,tplat2]),
           ([rtlon0,rtlon1,rtlon2],[rtlat0,rtlat1,rtlat2]),
           ([btlon0,btlon1,btlon2],[btlat0,btlat1,btlat2]),
           ([ltlon0,ltlon1,ltlon2],[ltlat0,ltlat1,ltlat2]),
        """
        g = Geod(ellps="WGS84")

        # Top of bounding box
        # g.npts returns a list of tuples of lon/lat pairs
        #    [(lon0,lat0),(lon1,lat1),(lon2,lat2)]
        # zip reformats that into 2 tuples of lons and lats
        #    [(lon0,lon1,lon2),(lat0,lat1,lat2)]
        # list(tplons) returns list of lons
        #       [lon0,lon1,lon2]
        # list(tplats) returns list of lats
        #       [lat0,lat1,lat2]
        tplons, tplats = zip(
            *g.npts(
                self.corners[0].lon,
                self.corners[0].lat,
                self.corners[1].lon,
                self.corners[1].lat,
                npts,
                radians=True,
            )
        )
        # Right side of bounding box
        rtlons, rtlats = zip(
            *g.npts(
                self.corners[1].lon,
                self.corners[1].lat,
                self.corners[2].lon,
                self.corners[2].lat,
                npts,
                radians=True,
            )
        )
        # Bottom of bounding box
        btlons, btlats = zip(
            *g.npts(
                self.corners[2].lon,
                self.corners[2].lat,
                self.corners[3].lon,
                self.corners[3].lat,
                npts,
                radians=True,
            )
        )
        # Left side of bounding box
        ltlons, ltlats = zip(
            *g.npts(
                self.corners[3].lon,
                self.corners[3].lat,
                self.corners[0].lon,
                self.corners[0].lat,
                npts,
                radians=True,
            )
        )

        retval = [
            (list(tplons), list(tplats)),
            (list(rtlons), list(rtlats)),
            (list(btlons), list(btlats)),
            (list(ltlons), list(ltlats)),
        ]
        return retval


class PlanarPolygonDefinition(CoordinateDefinition):
    """Planar polygon definition."""

    def __init__(self, lons, lats, nprocs=1):
        if lons.shape != lats.shape:
            raise ValueError("lon and lat arrays must have same shape")
        elif lons.ndim > 2:
            raise ValueError("Only 1 and 2 dimensional swaths are allowed")

        if lons.shape == lats.shape and lons.dtype == lats.dtype:
            self.shape = lons.shape
            self.size = lons.size
            self.ndim = lons.ndim
            self.dtype = lons.dtype
        else:
            raise ValueError(
                (
                    "%s must be created with either "
                    "lon/lats of the same shape with same dtype"
                )
                % self.__class__.__name__
            )

        lats_type = type(lats)
        lons_type = type(lons)

        if lons_type != lats_type:
            raise TypeError("lons and lats must be of same type")
        elif lons is not None:
            if lons.shape != lats.shape:
                raise ValueError("lons and lats must have same shape")

        self.nprocs = nprocs

        # check the latitutes
        if lats is not None and ((lats.min() < -90.0 or lats.max() > +90.0)):
            # throw exception
            raise ValueError("Some latitudes are outside the [-90.;+90] validity range")
        else:
            self.lats = lats

        # check the longitudes
        if lons is not None and ((lons.min() < -180.0 or lons.max() >= +180.0)):
            # issue warning
            warnings.warn(
                "All geometry objects expect longitudes in the [-180:+180[ range. "
                "We will now automatically wrap your longitudes into [-180:+180[, "
                "and continue. "
                "To avoid this warning next time, use routine utils.wrap_longitudes()."
            )
            # wrap longitudes to [-180;+180[
            self.lons = utils.wrap_longitudes(lons)
        else:
            self.lons = lons

        self.cartesian_coords = None

    def get_bounding_box_lonlats(self, npts=100):
        """Return array of lon/lats along the bounding lat/lon lines.

        Parameters
        ----------
        npts: int
            Number of points to return along each line

        Returns
        -------
        (top, right, bottom, left) : 4 tuples containing lists
                                    of len npts of lons/lats
        retval = (list(tplons),list(tplats)),
                 (list(rtlons),list(rtlats)),
                 (list(btlons),list(btlats)),
                 (list(ltlons),list(ltlats))

        eg for n=3
           ([tplon0,tplon1,tplon2],[tplat0,tplat1,tplat2]),
           ([rtlon0,rtlon1,rtlon2],[rtlat0,rtlat1,rtlat2]),
           ([btlon0,btlon1,btlon2],[btlat0,btlat1,btlat2]),
           ([ltlon0,ltlon1,ltlon2],[ltlat0,ltlat1,ltlat2]),
        """
        # Top of bounding box
        tplons = np.linspace(self.corners[0].lon, self.corners[1].lon, npts)
        tplats = np.linspace(self.corners[0].lat, self.corners[1].lat, npts)
        # Right side of bounding box
        rtlons = np.linspace(self.corners[1].lon, self.corners[2].lon, npts)
        rtlats = np.linspace(self.corners[1].lat, self.corners[2].lat, npts)
        # Bottom of bounding box
        btlons = np.linspace(self.corners[2].lon, self.corners[3].lon, npts)
        btlats = np.linspace(self.corners[2].lat, self.corners[3].lat, npts)
        # Left side of bounding box
        ltlons = np.linspace(self.corners[3].lon, self.corners[0].lon, npts)
        ltlats = np.linspace(self.corners[3].lat, self.corners[0].lat, npts)

        retval = [
            (list(tplons), list(tplats)),
            (list(rtlons), list(rtlats)),
            (list(btlons), list(btlats)),
            (list(ltlons), list(ltlats)),
        ]
        return retval

    @property
    def corners(self):
        """Return corners."""
        # print '    In 2D false corners for: '+str(self.name)
        try:
            # print '        Corners already set, returning'
            return super(CoordinateDefinition, self).corners
        except ValueError:
            pass

        return get_2d_false_corners(self)

    def __contains__(self, point):
        """Idenfity if point inside the 4 corners of the current area.

        This DOES NOT use spherical geometry / great circle arcs.
        """
        corners = self.corners

        if isinstance(point, tuple):
            from pyresample.spherical_geometry import Coordinate

            retval = planar_point_inside(Coordinate(*point), corners)
        else:
            retval = planar_point_inside(point, corners)

        # print '        retval from FALSE CORNERS contains '+str(retval)

        return retval

    def intersection(self, other):
        """Return current area intersection polygon corners against other.

        Parameters
        ----------
        other : object
            Instance of subclass of BaseDefinition

        Returns
        -------
        (corner1, corner2, corner3, corner4) : tuple of points
        """
        self_corners = self.corners

        other_corners = get_2d_false_corners(other)

        # shell()

        return planar_intersection_polygon(self_corners, other_corners)

    def overlaps_minmaxlatlon(self, other):
        """Determine if overlaps."""
        log.info("PlanarPolygonDefinition overlaps_minmaxlatlon")
        return self.overlaps(other)

    def overlaps(self, other):
        """Test if the current area overlaps the *other* area.

        This is based
        solely on the corners of areas, assuming the boundaries to be straight
        lines.

        Parameters
        ----------
        other : object
            Instance of subclass of BaseDefinition

        Returns
        -------
        overlaps : bool
        """
        self_corners = self.corners
        other_corners = get_2d_false_corners(other)

        log.info("    PlanarPolygon Overlaps Self False Corners: " + str(self_corners))
        log.info(
            "    PlanarPolygon Overlaps Other False Corners: " + str(other_corners)
        )

        # Previously just did if i in other or if i in self.
        # This does not take 2d_false_corners into account
        # when doing i in *area_definition* (because it uses
        # area_definition.__contains__, which does not use
        # planar_point_inside, but spherical point_inside.
        for i in self_corners:
            if planar_point_inside(i, other_corners):
                log.info("    Point " + str(i) + " in other")
                return True
        for i in other_corners:
            if planar_point_inside(i, self_corners):
                log.info("    Point " + str(i) + " in self")
                return True

        self_line1 = Line(self_corners[0], self_corners[1])
        self_line2 = Line(self_corners[1], self_corners[2])
        self_line3 = Line(self_corners[2], self_corners[3])
        self_line4 = Line(self_corners[3], self_corners[0])

        other_line1 = Line(other_corners[0], other_corners[1])
        other_line2 = Line(other_corners[1], other_corners[2])
        other_line3 = Line(other_corners[2], other_corners[3])
        other_line4 = Line(other_corners[3], other_corners[0])

        for i in (self_line1, self_line2, self_line3, self_line4):
            for j in (other_line1, other_line2, other_line3, other_line4):
                if i.intersects(j):
                    return True
        return False


class Line(object):
    """A Line between two lat/lon points."""

    start = None
    end = None

    def __init__(self, start, end):
        """Initialize Line."""
        self.start, self.end = start, end

    def __eq__(self, other):
        """Test two lines equal."""
        if (
            abs(self.start.lon - other.start.lon) < EPSILON
            and abs(self.end.lon - other.end.lon) < EPSILON
            and abs(self.start.lat - other.start.lat) < EPSILON
            and abs(self.end.lat - other.end.lat) < EPSILON
        ):
            return 1
        return 0

    def __ne__(self, other):
        """Test two lines not equal."""
        return not self.__eq__(other)

    def __str__(self):
        """Return string."""
        return str(self.start) + " -> " + str(self.end)

    def __repr__(self):
        """Return representation."""
        return str(self.start) + " -> " + str(self.end)

    def intersects(self, other_line):
        """Test two lines intersect.

        Says if two lines defined by the current line and the *other_line*
        intersect. A line is defined as the shortest tracks between two points.
        """
        intpt = self.intersection(other_line)
        return bool(intpt)

    def intersection(self, other):
        """Identify intersection between two lines.

        Says where, if two lines defined by the current line and the
        *other_line* intersect.
        """
        log.info("self: " + str(self) + " other: " + str(other))
        if self == other:
            # Used to be return True, that is definitely not right (expects Coordinate)
            # Do we want start or end ? Does it matter? Lines are the same, everything
            # is an intersection.
            return self.start
        # If any of the start/end points match, return that point.
        if self.end == other.start or self.end == other.end:
            return self.end
        if self.start == other.start or self.start == other.end:
            return self.start

        # Line equation: y = mx + b
        # m = (y2-y1)/(x2-x1)
        # B_self = y - M_self*x
        # Pick any x/y on the line - try end point
        # B_self = self.end.lat - M_self*self.end.lon
        # B_other = other.end.lat - M_self*self.end.lon
        from pyresample.spherical_geometry import Coordinate

        selfendlon = self.end.lon
        selfstartlon = self.start.lon
        otherendlon = other.end.lon
        otherstartlon = other.start.lon
        # Not sure if this is necessary, or good...
        #        if self.end.lon < 0:
        #            selfendlon = self.end.lon + 2*math.pi
        #        if self.start.lon < 0:
        #            selfstartlon = self.start.lon + 2*math.pi
        #        if other.end.lon < 0:
        #            otherendlon = other.end.lon + 2*math.pi
        #        if other.start.lon < 0:
        #            otherstartlon = other.start.lon + 2*math.pi

        log.info(
            "    self lons: "
            + str(math.degrees(selfstartlon))
            + " "
            + str(math.degrees(selfendlon))
            + " other lons: "
            + str(math.degrees(otherstartlon))
            + " "
            + str(math.degrees(otherendlon))
        )

        # If both vertical, will be no intersection
        if (
            abs(selfendlon - selfstartlon) < EPSILON
            and abs(otherendlon - otherstartlon) < EPSILON
        ):
            log.info("    Both vertical, no intersection")
            return None
        # If self is vertical, but not parallel, intersection will be selfstartlon
        # and lat = Mother*lon+B_other
        if abs(selfendlon - selfstartlon) < EPSILON:
            lon = selfstartlon
            M_other = (other.end.lat - other.start.lat) / (otherendlon - otherstartlon)
            B_other = other.end.lat - M_other * otherendlon
            lat = M_other * lon + B_other
            log.info("    self is vertical")
            # Make sure it falls within the segment and not outside.
            # Previously was only checking lat, need to
            # also check lon or opposite side of world would match
            if (
                lat > min([self.end.lat, self.start.lat])
                and lat < max([self.end.lat, self.start.lat])
                and lon > min([otherendlon, otherstartlon])
                and lon < max([otherendlon, otherstartlon])
            ):
                log.info("        and intersects")
                # Apparently Coordinate takes degrees ??? And must be -180 to 180 ?!
                # MLS use wrap_longitudes?
                if lon > math.pi:
                    lon -= 2 * math.pi
                return Coordinate(math.degrees(lon), math.degrees(lat))
            else:
                return None
        # same for other
        if abs(otherendlon - otherstartlon) < EPSILON:
            lon = otherstartlon
            M_self = (self.end.lat - self.start.lat) / (selfendlon - selfstartlon)
            B_self = self.end.lat - M_self * selfendlon
            lat = M_self * lon + B_self
            log.info("    other is vertical")
            # Make sure it falls within the segment and not outside.
            # Previously was only checking lat, need to
            # also check lon or opposite side of world would match
            if (
                lat > min([other.end.lat, other.start.lat])
                and lat < max([other.end.lat, other.start.lat])
                and lon > min([selfendlon, selfstartlon])
                and lon < max([selfendlon, selfstartlon])
            ):
                log.info("        and intersects")
                # Apparently Coordinate takes degrees ??? And must be -180 to 180 ?!
                # MLS Use wrap_longitudes?
                if lon > math.pi:
                    lon -= 2 * math.pi
                return Coordinate(math.degrees(lon), math.degrees(lat))
            else:
                return None

        # Get slopes of the lines
        M_self = (self.end.lat - self.start.lat) / (selfendlon - selfstartlon)
        M_other = (other.end.lat - other.start.lat) / (otherendlon - otherstartlon)

        # If they are parallel, no intersection
        if (M_self - M_other) < EPSILON:
            log.info("    self and other are parallel, no intersection")
            return None

        # Get the y-intercepts of the lines
        B_self = self.end.lat - M_self * selfendlon
        B_other = other.end.lat - M_other * otherendlon

        # Solve the equation
        # y=m1x+b1 and y=m2x+b2, equate y's so m1x+b1=m2x+b2, x = (b1-b2)/(m2-m1)
        # equate x's so x=(y-b1)/m1=(y-b2)/m2, y = (b1m2-b2m1)/(m2-m1)
        lon = (B_self - B_other) / (M_other - M_self)
        lat = (B_self * M_other - B_other * M_self) / (M_other - M_self)

        # Make sure lat/lon intersects within the line segment, and not outside.
        if (
            lat > min([other.end.lat, other.start.lat])
            and lat < max([other.end.lat, other.start.lat])
            and lon > min([otherendlon, otherstartlon])
            and lon < max([otherendlon, otherstartlon])
            and lat > min([self.end.lat, self.start.lat])
            and lat < max([self.end.lat, self.start.lat])
            and lon > min([selfendlon, selfstartlon])
            and lon < max([selfendlon, selfstartlon])
        ):
            log.info("    self and other intersect within segment")
            # Apparently Coordinate takes degrees ??? And must be -180 to 180 ?!
            # MLS use wrap longitudes?
            if lon > math.pi:
                lon -= 2 * math.pi
            return Coordinate(math.degrees(lon), math.degrees(lat))
        else:
            log.info("    self and other intersect, but not within segment")
            return None


def get_2d_false_corners(box_def):
    """Identify false corners."""
    # print '    In 2D false corners for: '+str(box_def.name)

    min_row = 0
    max_row = -1
    min_col = 0
    max_col = -1
    side1 = box_def.get_lonlats(data_slice=(min_row, slice(None)))
    side2 = box_def.get_lonlats(data_slice=(slice(None), max_col))
    side3 = box_def.get_lonlats(data_slice=(max_row, slice(None)))
    side4 = box_def.get_lonlats(data_slice=(slice(None), min_col))

    tries = 0
    while (
        tries < 500
        and np.ma.count(box_def.get_lonlats(data_slice=(min_row, slice(None)))[1]) < 10
    ):
        min_row += 1
        tries += 1
    if tries:
        side1 = box_def.get_lonlats(data_slice=(min_row + 1, slice(None)))
        log.info(
            "Needed some data in side 1, incremented slice number "
            + str(tries)
            + " times. Now have "
            + str(np.ma.count(side1[1]))
            + " valid of "
            + str(np.ma.count(side1[1].mask))
        )

    tries = 0
    while (
        tries < 500
        and np.ma.count(box_def.get_lonlats(data_slice=(slice(None), max_col))[0]) < 10
    ):
        max_col -= 1
        tries += 1
    if tries:
        side2 = box_def.get_lonlats(data_slice=(slice(None), max_col - 1))
        log.info(
            "Needed some data in side 2, decremented slice number "
            + str(tries)
            + " times. Now have "
            + str(np.ma.count(side2[0]))
            + " valid of "
            + str(np.ma.count(side2[0].mask))
        )

    tries = 0
    while (
        tries < 500
        and np.ma.count(box_def.get_lonlats(data_slice=(max_row, slice(None)))[0]) < 10
    ):
        max_row -= 1
        tries += 1
    if tries:
        side3 = box_def.get_lonlats(data_slice=(max_row - 1, slice(None)))
        log.info(
            "Needed some data in side 3, decremented slice number "
            + str(tries)
            + " times. Now have "
            + str(np.ma.count(side3[0]))
            + " valid of "
            + str(np.ma.count(side3[0].mask))
        )

    tries = 0
    while (
        tries < 500
        and np.ma.count(box_def.get_lonlats(data_slice=(slice(None), min_col))[1]) < 10
    ):
        min_col += 1
        tries += 1
    if tries:
        side4 = box_def.get_lonlats(data_slice=(slice(None), min_col + 1))
        log.info(
            "Needed some data in side 4, incremented slice number "
            + str(tries)
            + " times. Now have "
            + str(np.ma.count(side4[1]))
            + " valid of "
            + str(np.ma.count(side4[1].mask))
        )

    # shell()

    # These all need to maintain mask.
    selflons = np.ma.concatenate((side1[0], side2[0], side3[0], side4[0]))
    selflons = np.ma.where(selflons < 0, selflons + 360, selflons)
    # MLS use wrap_longitudes? Figure out prime meridian vs dateline...
    # if side4[0].min() > side2[0].max():
    #    selflons = np.ma.where(selflons<0,selflons+360,selflons)
    selflats = np.ma.concatenate((side1[1], side2[1], side3[1], side4[1]))

    # self_corners = self.corners
    # other_corners = other.corners
    minlon = selflons.min()
    maxlon = selflons.max()
    # MLS use wrap_longitudes?
    if minlon > 180:
        minlon -= 360
    if maxlon > 180:
        maxlon -= 360
    minlat = selflats.min()
    maxlat = selflats.max()

    # print 'IN PlanarPolygonDefinition CORNERS for '+box_def.name+\
    #    ' min/max lat min/max lon:'+\
    #    str(minlat)+' '+str(maxlat)+' '+str(minlon)+' '+str(maxlon)

    from pyresample.spherical_geometry import Coordinate

    return [
        Coordinate(minlon, maxlat),
        Coordinate(maxlon, maxlat),
        Coordinate(maxlon, minlat),
        Coordinate(minlon, minlat),
    ]


def planar_intersection_polygon(area_corners, segment_corners):
    """Get the intersection polygon between two areas."""
    # First test each
    lons = np.array([])
    lats = np.array([])
    for segment_corner in segment_corners:
        if planar_point_inside(segment_corner, area_corners):
            currlon = segment_corner.lon
            # MLS use wrap_longitudes?
            if currlon < 0:
                currlon += 2 * math.pi
            lons = np.concatenate((lons, [currlon]))
            lats = np.concatenate((lats, [segment_corner.lat]))
            log.info("Adding intersection from segment " + str(segment_corner))
    for area_corner in area_corners:
        if planar_point_inside(area_corner, segment_corners):
            currlon = area_corner.lon
            # MLS use wrap_longitudes?
            if currlon < 0:
                currlon += 2 * math.pi
            lons = np.concatenate((lons, [currlon]))
            lats = np.concatenate((lats, [area_corner.lat]))
            log.info("Adding intersection from area " + str(area_corner))

    area_line1 = Line(area_corners[0], area_corners[1])
    area_line2 = Line(area_corners[1], area_corners[2])
    area_line3 = Line(area_corners[2], area_corners[3])
    area_line4 = Line(area_corners[3], area_corners[0])

    segment_line1 = Line(segment_corners[0], segment_corners[1])
    segment_line2 = Line(segment_corners[1], segment_corners[2])
    segment_line3 = Line(segment_corners[2], segment_corners[3])
    segment_line4 = Line(segment_corners[3], segment_corners[0])

    for i in (area_line1, area_line2, area_line3, area_line4):
        for j in (segment_line1, segment_line2, segment_line3, segment_line4):
            intersect = i.intersection(j)
            if intersect:
                log.info("Adding actual intersection " + str(intersect))
                currlon = intersect.lon
                # MLS use wrap_longitudes?
                if intersect.lon < 0:
                    currlon += 2 * math.pi
                lons = np.concatenate((lons, [currlon]))
                lats = np.concatenate((lats, [intersect.lat]))

    minlon = math.degrees(lons.min())
    maxlon = math.degrees(lons.max())
    minlat = math.degrees(lats.min())
    maxlat = math.degrees(lats.max())
    # Coordinate MUST be between -180 and 180
    # MLS use wrap_longitudes?
    if minlon > 180:
        minlon -= 180
    if maxlon > 180:
        maxlon -= 180
    from pyresample.spherical_geometry import Coordinate

    return [
        Coordinate(minlon, maxlat),
        Coordinate(maxlon, maxlat),
        Coordinate(maxlon, minlat),
        Coordinate(minlon, minlat),
    ]


#    for seg_pt in seg_pts_in_area:
#
#
#
#
#
# def planar_point_inside(point, boxdef):
def planar_point_inside(point, corners):
    """Identify point inside 4 corners.

    This DOES NOT USE great circle arcs as area
    boundaries.
    """
    #    lons = boxdef.get_lonlats()[0]
    lons = np.ma.array([corn.lon for corn in corners])
    lats = np.ma.array([corn.lat for corn in corners])
    # MLS use wrap_longitudes?
    lons = np.ma.where(lons < 0, lons + 2 * math.pi, lons)

    #    lats = boxdef.get_lonlats()[1]
    #    corners = boxdef.corners
    minlon = lons.min()
    maxlon = lons.max()
    minlat = lats.min()
    maxlat = lats.max()
    # MLS use wrap_longitudes?
    if point.lon < 0:
        point.lon += 2 * math.pi
    #    print '    IN PlanarPolygonDefinition point_inside!!! '+\
    #        ' point: '+str(point)+' '+str(math.degrees(minlat))+' '+
    #                 str(math.degrees(maxlat))+' '+str(math.degrees(minlon))+' '+
    #                 str(math.degrees(maxlon))
    #        ' point: '+str(point)+'\n'+\
    #        'c0 '+str(corners[0])+'\n'+\
    #        'c1 '+str(corners[1])+'\n'+\
    #        'c2 '+str(corners[2])+'\n'+\
    #        'c3 '+str(corners[3])+'\n'+\
    #        str(minlat)+' '+str(maxlat)+' '+str(minlon)+' '+str(maxlon)
    # MLS 20160405 NOTE point prints degrees for str, but point.lon and point.lat are
    # stored as radians.
    # minlon/maxlon also radians.
    # This is why big sectors were failing, after fixing the
    # "other side of the world" problem.
    # Also, Coordinate takes degrees when passing it a lat lon
    if minlon < point.lon < maxlon and minlat < point.lat < maxlat:
        return True
    return False


def _get_slice(segments, shape):
    """Segment a 1D or 2D array."""
    if not (1 <= len(shape) <= 2):
        raise ValueError("Cannot segment array of shape: %s" % str(shape))
    else:
        size = shape[0]
        slice_length = np.ceil(float(size) / segments)
        start_idx = 0
        end_idx = slice_length
        while start_idx < size:
            if len(shape) == 1:
                yield slice(start_idx, end_idx)
            else:
                yield (slice(start_idx, end_idx), slice(None))
            start_idx = end_idx
            end_idx = min(start_idx + slice_length, size)


def _flatten_cartesian_coords(cartesian_coords):
    """Flatten array to (n, 3) shape."""
    shape = cartesian_coords.shape
    if len(shape) > 2:
        cartesian_coords = cartesian_coords.reshape(shape[0] * shape[1], 3)
    return cartesian_coords


def _get_highest_level_class(obj1, obj2):
    """Get highest level class."""
    if not issubclass(obj1.__class__, obj2.__class__) or not issubclass(
        obj2.__class__, obj1.__class__
    ):
        raise TypeError(
            "No common superclass for %s and %s" % (obj1.__class__, obj2.__class__)
        )

    if obj1.__class__ == obj2.__class__:
        klass = obj1.__class__
    elif issubclass(obj1.__class__, obj2.__class__):
        klass = obj2.__class__
    else:
        klass = obj1.__class__
    return klass



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Read Advanced Microwave Scanning Radiometer (AMSR2) data products."""

# Python Standard Libraries
from glob import glob
import logging
from os.path import basename

# Third-Party Libraries
import numpy
import pandas
import xarray

LOG = logging.getLogger(__name__)

varnames = {
    "Brightness_Temperature_10_GHzH": "tb10h",
    "Brightness_Temperature_10_GHzV": "tb10v",
    "Brightness_Temperature_18_GHzH": "tb18h",
    "Brightness_Temperature_18_GHzV": "tb18v",
    "Brightness_Temperature_23_GHzH": "tb23h",
    "Brightness_Temperature_23_GHzV": "tb23v",
    "Brightness_Temperature_36_GHzH": "tb36h",
    "Brightness_Temperature_36_GHzV": "tb36v",
    "Brightness_Temperature_6_GHzH": "tb6h",
    "Brightness_Temperature_6_GHzV": "tb6v",
    "Brightness_Temperature_7_GHzH": "tb7h",
    "Brightness_Temperature_7_GHzV": "tb7v",
    "Brightness_Temperature_89_GHz_AH": "tb89hA",
    "Brightness_Temperature_89_GHz_AV": "tb89vA",
    "Brightness_Temperature_89_GHz_BH": "tb89hB",
    "Brightness_Temperature_89_GHz_BV": "tb89hB",
}
land_num = {"6": 0, "7": 1, "10": 2, "18": 3, "23": 4, "36": 5, "89A": 0, "89B": 1}
land_var = {
    "6": "Land_Ocean_Flag_6_to_36",
    "7": "Land_Ocean_Flag_6_to_36",
    "10": "Land_Ocean_Flag_6_to_36",
    "18": "Land_Ocean_Flag_6_to_36",
    "23": "Land_Ocean_Flag_6_to_36",
    "36": "Land_Ocean_Flag_6_to_36",
    "89A": "Land_Ocean_Flag_89",
    "89B": "Land_Ocean_Flag_89",
}
chan_nums = {
    "Brightness_Temperature_6_GHzV": 1,
    "Brightness_Temperature_6_GHzH": 2,
    "Brightness_Temperature_7_GHzV": 3,
    "Brightness_Temperature_7_GHzH": 4,
    "Brightness_Temperature_10_GHzV": 5,
    "Brightness_Temperature_10_GHzH": 6,
    "Brightness_Temperature_18_GHzV": 7,
    "Brightness_Temperature_18_GHzH": 8,
    "Brightness_Temperature_23_GHzV": 9,
    "Brightness_Temperature_23_GHzH": 10,
    "Brightness_Temperature_36_GHzV": 11,
    "Brightness_Temperature_36_GHzH": 12,
    "Brightness_Temperature_89_GHz_AV": 13,
    "Brightness_Temperature_89_GHz_AH": 14,
    "Brightness_Temperature_89_GHz_BV": 13,
    "Brightness_Temperature_89_GHz_BH": 14,
}

interface = "readers"
family = "standard"
name = "amsr2_netcdf"
source_names = ["amsr2"]


def read_amsr2_winds(wind_xarray):
    """Reformat AMSR2 xarray object appropriately.

    * variables: latitude, longitude, time, wind_speed_kts
    * attributes: source_name, platform_name, data_provider,
      interpolation_radius_of_influence
    """
    MS_TO_KTS = 1.94384
    LOG.info("Reading AMSR2 data")
    # Set attributes appropriately
    wind_xarray.attrs["source_name"] = "amsr2"
    wind_xarray.attrs["platform_name"] = "gcom-w1"
    wind_xarray.attrs["interpolation_radius_of_influence"] = 10000
    # AMSR2 is one text file per datafile, so don't append
    wind_xarray.attrs["overwrite_text_file"] = True
    wind_xarray.attrs["append_text_file"] = False
    # https://www.ospo.noaa.gov/Products/atmosphere/gpds/about_amsr2.html
    # OCEAN Winds based on 37GHz, which is 7km x 12km ground resolution
    wind_xarray.attrs["sample_distance_km"] = 7.0
    if "creator_name" in wind_xarray.attrs and "NOAA" in wind_xarray.creator_name:
        wind_xarray.attrs["data_provider"] = "star"

    # Set lat/lons appropriately
    wind_xarray = wind_xarray.rename(
        {
            "Latitude_for_Low_Resolution": "latitude",
            "Longitude_for_Low_Resolution": "longitude",
        }
    )
    wind_xarray = wind_xarray.set_coords(["latitude", "longitude"])
    wind_xarray = wind_xarray.reset_coords(
        ["Latitude_for_High_Resolution", "Longitude_for_High_Resolution"]
    )

    # Set wind_speed_kts appropriately
    # convert to kts
    wind_xarray["wind_speed_kts"] = wind_xarray["WSPD"] * MS_TO_KTS

    # # Only keep the good wind speeds
    # wind_xarray['wind_speed_kts'] = xarray.where(wind_xarray['WSPD_QC'] == 0,
    # wind_xarray['wind_speed_kts'], numpy.nan)

    wind_xarray["wind_speed_kts"].attrs = wind_xarray["WSPD"].attrs
    wind_xarray["wind_speed_kts"].attrs["units"] = "kts"
    wind_xarray["wind_speed_kts"] = wind_xarray["wind_speed_kts"].assign_coords(
        latitude=wind_xarray.latitude, longitude=wind_xarray.longitude
    )

    # Set time array appropriately

    dtstrs = []
    LOG.info("Reading scan_times")
    for scan_time in wind_xarray["Scan_Time"]:
        dtstrs += [
            "{0:04.0f}{1:02.0f}{2:02.0f}T{3:02.0f}{4:02.0f}{5:02.0f}".format(
                *tuple([xx for xx in scan_time.values])
            )
        ]
    # Have to set it on the actual xarray so it becomes a xarray format time series
    # (otherwise if you set it directly to ts, it is a pandas format time series, and
    # expand_dims doesn't exist).
    time_array = pandas.to_datetime(
        dtstrs, format="%Y%m%dT%H%M%S", errors="coerce"
    ).tolist()
    LOG.info("Setting list of times")
    tss = [time_array for ii in range(0, wind_xarray["wind_speed_kts"].shape[1])]
    LOG.info("Setting time DataArray")
    wind_xarray["time"] = xarray.DataArray(
        data=numpy.array(tss).transpose(),
        coords=wind_xarray.wind_speed_kts.coords,
        name="time",
    )
    wind_xarray = wind_xarray.set_coords(["time"])
    return {"WINDS": wind_xarray}


def read_amsr2_mbt(full_xarray, varname, time_array=None):
    """
    Reformat AMSR2 xarray object appropriately.

    * variables: latitude, longitude, time,
      brightness temperature variables
    * attributes: source_name, platform_name, data_provider,
      interpolation_radius_of_influence
    """
    LOG.info("Reading AMSR data %s", varname)
    sub_xarray = xarray.Dataset()
    sub_xarray.attrs = full_xarray.attrs.copy()
    # Set attributes appropriately

    # Mappings are Brightness_Temperature_89_GHz_AH -> Latitude_for_89A, etc
    # Mappings are Brightness_Temperature_10_GHzH -> Latitude_for_10, etc
    chanstr = varname.replace("Brightness_Temperature_", "")
    chanstr = chanstr.replace("_GHz_", "")
    chanstr = chanstr.replace("_GHz", "")
    chanstr = chanstr.replace("H", "")
    chanstr = chanstr.replace("V", "")

    # Set lat/lons appropriately
    sub_xarray["latitude"] = full_xarray["Latitude_for_{0}".format(chanstr)]
    sub_xarray["longitude"] = full_xarray["Longitude_for_{0}".format(chanstr)]
    sub_xarray[varnames[varname]] = full_xarray[varname]
    sub_xarray[varnames[varname]].attrs["channel_number"] = chan_nums[varname]
    sub_xarray.set_coords(["latitude", "longitude"])

    # https://www.ospo.noaa.gov/Products/atmosphere/gpds/about_amsr2.html
    # 37GHz, 7km x 12km ground resolution
    # 89GHz, 3km x 5km ground resolution

    # MTIFs need to be "prettier" for PMW products, so 2km resolution for all channels
    # sub_xarray.attrs['sample_distance_km'] = 3.0
    sub_xarray.attrs["sample_distance_km"] = 2.0
    sub_xarray.attrs["interpolation_radius_of_influence"] = 10000
    for dim in sub_xarray.dims.keys():
        if "low_rez" in dim:
            # MTIFs need to be "prettier" for PMW products, so 2km resolution for all
            # channels. sub_xarray.attrs['sample_distance_km'] = 7.0
            sub_xarray.attrs["sample_distance_km"] = 2.0
            sub_xarray.attrs["interpolation_radius_of_influence"] = 20000

    # See dictionaries above for appropriate land mask array locations for each variable
    full_xarray["LandMask"] = xarray.DataArray(
        full_xarray[land_var[chanstr]].to_masked_array()[land_num[chanstr], :, :],
        coords=full_xarray[varname].coords,
    )

    if time_array is None:

        # Set time appropriately
        dtstrs = []
        LOG.info("Reading scan_times, for dims %s", sub_xarray[varnames[varname]].dims)
        for scan_time in full_xarray["Scan_Time"]:
            dtstrs += [
                "{0:04.0f}{1:02.0f}{2:02.0f}T{3:02.0f}{4:02.0f}{5:02.0f}".format(
                    *tuple([xx for xx in scan_time.values])
                )
            ]
        # Have to set it on the actual xarray so it becomes a xarray format time series
        # (otherwise if you set it directly to ts, it is a pandas format time series,
        # and expand_dims doesn't exist).
        curr_time_array = pandas.to_datetime(
            dtstrs, format="%Y%m%dT%H%M%S", errors="coerce"
        ).tolist()
        LOG.info("    Setting list of times")
        tss = [
            curr_time_array for ii in range(0, sub_xarray[varnames[varname]].shape[1])
        ]
        LOG.info("    Setting time DataArray")
        sub_xarray["time"] = xarray.DataArray(
            data=numpy.array(tss).transpose(),
            coords=full_xarray[varname].coords,
            name="time",
        )
        sub_xarray = sub_xarray.set_coords(["time"])
    else:
        LOG.info(
            "Using existing scan_times for dims %s", sub_xarray[varnames[varname]].dims
        )
        sub_xarray["time"] = time_array
    from geoips.xarray_utils.time import (
        get_min_from_xarray_time,
        get_max_from_xarray_time,
    )

    sub_xarray.attrs["start_datetime"] = get_min_from_xarray_time(sub_xarray, "time")
    sub_xarray.attrs["end_datetime"] = get_max_from_xarray_time(sub_xarray, "time")
    return sub_xarray


def read_amsr2_data(full_xarray, chans):
    """Read non-AMSR2_OCEAN data."""
    full_xarray = full_xarray.reset_coords(full_xarray.coords)

    sunzen = 90 - full_xarray.Sun_Elevation
    satzen = full_xarray.Earth_Incidence_Angle
    satazm = full_xarray.Earth_Azimuth_Angle
    loqf = full_xarray.Pixel_Data_Quality_6_to_36
    hiqf = full_xarray.Pixel_Data_Quality_89
    sunglint_flag = full_xarray.Sun_Glint_Flag
    sensor_scan_angle = full_xarray.Scan_Angle
    xarrays = {}
    # Every single channel has a different set of Lat/Lons!
    for varname in full_xarray.variables:
        if chans is not None and varname not in chans:
            LOG.info("SKIPPING: Variable %s not requested in %s", varname, chans)
        if "Brightness" in varname:
            usetime = None
            for xra in list(xarrays.values()):
                if xra.time.dims == full_xarray[varname].dims:
                    usetime = xra.time
            new_xarray = read_amsr2_mbt(full_xarray, varname, usetime)
            if sunzen.dims == tuple(new_xarray.dims):
                new_xarray["satellite_azimuth_angle"] = satazm
                new_xarray["satellite_zenith_angle"] = satzen
                new_xarray["solar_zenith_angle"] = sunzen
                new_xarray["QualityFlag"] = loqf
                new_xarray["SunGlintFlag"] = sunglint_flag
                new_xarray["sensor_scan_angle"] = sensor_scan_angle
            elif hiqf.dims == tuple(new_xarray.dims):
                new_xarray["QualityFlag"] = hiqf
            xarrays[varname] = new_xarray
        else:
            LOG.info("SKIPPING variable %s", varname)

    return xarrays


def call(
    fnames,
    metadata_only=False,
    chans=None,
    area_def=None,
    self_register=False,
    test_arg="AMSR2 Default Test Arg",
):
    """
    Read AMSR2 netcdf data products.

    Parameters
    ----------
    fnames : list
        * List of strings, full paths to files
    metadata_only : bool, default=False
        * Return before actually reading data if True
    chans : list of str, default=None
        * List of desired channels (skip unneeded variables as needed).
        * Include all channels if None.
    area_def : pyresample.AreaDefinition, default=None
        * NOT YET IMPLEMENTED
        * Specify region to read
        * Read all data if None.
    self_register : str or bool, default=False
        * NOT YET IMPLEMENTED
        * register all data to the specified dataset id (as specified in the
          return dictionary keys).
        * Read multiple resolutions of data if False.

    Returns
    -------
    dict of xarray.Datasets
        * dictionary of xarray.Dataset objects with required Variables and
          Attributes.
        * Dictionary keys can be any descriptive dataset ids.

    See Also
    --------
    :ref:`xarray_standards`
        Additional information regarding required attributes and variables
        for GeoIPS-formatted xarray Datasets.
    """
    LOG.debug("AMSR2 reader test_arg: %s", test_arg)
    ingested = []
    for fname in fnames:
        # full_xarray = xarray.open_dataset(str(fname))
        full_xarrays = [xarray.open_dataset(str(x)) for x in fnames]
        full_xarray = xarray.merge(full_xarrays)
        full_xarray.attrs["data_provider"] = "unknown"
        full_xarray.attrs["source_file_names"] = [basename(fname)]
        full_xarray.attrs["source_name"] = "amsr2"
        full_xarray.attrs["platform_name"] = "gcom-w1"
        full_xarray.attrs["interpolation_radius_of_influence"] = 10000
        if "creator_name" in full_xarray.attrs and "NOAA" in full_xarray.creator_name:
            full_xarray.attrs["data_provider"] = "star"
        full_xarray.attrs["minimum_coverage"] = 20
        LOG.info("Read data from %s", fname)
        if metadata_only is True:
            from datetime import datetime

            full_xarray.attrs["start_datetime"] = datetime.strptime(
                full_xarray.attrs["time_coverage_start"][0:19], "%Y-%m-%dT%H:%M:%S"
            )
            full_xarray.attrs["end_datetime"] = datetime.strptime(
                full_xarray.attrs["time_coverage_end"][0:19], "%Y-%m-%dT%H:%M:%S"
            )
            LOG.info("metadata_only requested, returning without readind data")
            return {"METADATA": full_xarray}

        if hasattr(full_xarray, "title") and "AMSR2_OCEAN" in full_xarray.title:
            xarrays = read_amsr2_winds(full_xarray)

        elif hasattr(full_xarray, "title") and "MBT" in full_xarray.title:
            xarrays = read_amsr2_data(full_xarray, chans)

        elif hasattr(full_xarray, "title") and "PRECIP" in full_xarray.title:
            xarrays = read_amsr2_data(full_xarray, chans)
        ingested.append(xarrays)

    # Merge all datasets together:
    final_xarrays = {}
    for xr_key in xarrays.keys():
        xars = [xar[xr_key] for xar in ingested]
        final_xarrays[xr_key] = xarray.concat(xars, dim="Number_of_Scans")

    for dsname, curr_xarray in final_xarrays.items():
        LOG.info("Setting standard metadata")
        from geoips.xarray_utils.time import (
            get_min_from_xarray_time,
            get_max_from_xarray_time,
        )

        curr_xarray.attrs["start_datetime"] = get_min_from_xarray_time(
            curr_xarray, "time"
        )
        curr_xarray.attrs["end_datetime"] = get_max_from_xarray_time(
            curr_xarray, "time"
        )
    final_xarrays["METADATA"] = list(final_xarrays.values())[0][[]]
    return final_xarrays


def get_test_files(test_data_dir):
    """Generate test files for unit testing reader."""
    filepath = test_data_dir + "/test_data_amsr2/data/AMSR2-MBT*.nc"
    filelist = glob(filepath)
    tmp_xr = call(filelist)

    return tmp_xr


def get_test_parameters():
    """Generate a data key for unit testing."""
    return [{"data_key": "Brightness_Temperature_10_GHzH", "data_var": "tb10h"}]



# # # This source code is protected under the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""GeoIPS order-based procflow pydantic models init file."""

import logging
import importlib
import inspect
import pkgutil
import sys


LOG = logging.getLogger(__name__)


def collect_modules():
    """Dynamically find and import all submodules within a package."""
    modules = {}

    # Get the current package (i.e., geoips.pydantic)
    package = sys.modules[__name__]

    for _, module_name, is_pkg in pkgutil.walk_packages(  # NOQA
        package.__path__, package.__name__ + "."
    ):
        try:
            module = importlib.import_module(module_name)
            modules[module_name] = module
        except ImportError as e:
            LOG.warning(f"Could not import {module_name}: {e}")

    return modules


def collect_classes(modules):
    """Extract all classes from the given modules.

    Parameters
    ----------
    modules: dict
        - A dictionary of mod_name: module objects found within geoips.pydantic
    """
    classes = {}

    for module_name, module in modules.items():
        module_classes = {
            name: cls
            for name, cls in inspect.getmembers(module, inspect.isclass)
            if cls.__module__
            == module_name  # Ensure it's defined in the module, not imported
        }
        if module_classes:
            classes[module_name] = module_classes

    return classes


_modules = collect_modules()
_classes = collect_classes(_modules)  # NOQA



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Geotiff image rasterio-based output format."""

import logging

# Internal utilities
from geoips.interfaces import output_formatters

LOG = logging.getLogger(__name__)

interface = "output_formatters"
family = "image"
name = "cogeotiff_rgba"


def call(
    area_def,
    xarray_obj,
    product_name,
    output_fnames,
    product_name_title=None,
    mpl_colors_info=None,
    existing_image=None,
):
    """Create standard geotiff output using rasterio."""
    cogeotiff = output_formatters.get_plugin("cogeotiff")
    output_fnames = cogeotiff(
        area_def,
        xarray_obj,
        product_name,
        output_fnames,
        product_name_title,
        mpl_colors_info,
        existing_image,
    )

    return output_fnames



"""Processing workflow for order based data source processing."""

# Python Standard Libraries
from argparse import ArgumentParser
import logging

# GeoIPS imports
from geoips import interfaces
from geoips.pydantic.workflows import WorkflowPluginModel

LOG = logging.getLogger(__name__)

interface = "procflows"
family = "standard"
name = "order_based"


def call(workflow, fnames, command_line_args=None):
    """Run the order based procflow (OBP).

    Process the specified input data files using the OBP in the order of steps
    listed in the workflow definition file.

    Parameters
    ----------
    workflow: str
        The name of the workflow to process.
    fnames : list of str
        List of filenames from which to read data.
    command_line_args : list of str, None
        Command line arguments to pass to the workflow.
    """
    LOG.interactive(f"Begin processing '{workflow}' workflow.")
    wf_plugin = interfaces.workflows.get_plugin(workflow)
    wf = WorkflowPluginModel(**wf_plugin)

    handled_interfaces = ["readers"]
    for step_id, step_def in wf.spec.steps.items():
        interface = step_def.kind + "s"

        if interface not in handled_interfaces:
            LOG.interactive(
                "⚠️ Skipping unhandled interface '%s'. Would have called the '%s'"
                "plugin.",
                interface,
                step_def.name,
            )
            continue
        else:
            plg = getattr(interfaces, interface, None).get_plugin(step_def.name)
            LOG.interactive(
                "Beginning Step: '%s', plugin_kind: '%s', plugin_name:'%s'.",
                step_id,
                step_def.kind,
                step_def.name,
            )
            LOG.info("Arguments: '%s'", step_def.arguments)

            if interface == "readers":
                # TEMPORARY FIX: Remove when all readers are updated to accept
                # "variables"
                if "variables" in step_def.arguments:
                    step_def.arguments["chans"] = step_def.arguments.pop("variables")
                data = plg(fnames, **step_def.arguments)
                print(data)
            else:
                data = plg(data, **step_def.arguments)
            LOG.interactive(
                "Completed Step: step_id: '%s', plugin_kind: '%s', plugin_name: '%s'.",
                step_id,
                step_def.name,
                step_def.kind,
            )

    LOG.interactive(f"\nThe workflow '{workflow}' has finished processing.\n")


if __name__ == "__main__":

    parser = ArgumentParser(description="order-based procflow processing")
    parser.add_argument("workflow", help="The workflow name to process.")
    parser.add_argument("fnames", nargs="+", help="The filenames to process.")
    args = parser.parse_args()
    call(args.workflow, args.fnames)



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Unit test for retrieving commandline instructions for the GeoIPS CLI."""

from os import listdir, remove
from os.path import dirname, exists
import pytest

from geoips.commandline.cmd_instructions import get_instructions
from geoips.commandline.commandline_interface import GeoipsCLI


instruct_dir = f"{str(dirname(__file__))}/cmd_instructions"


def generate_id(dir_name):
    """Generate a test id for the instructions directory being tested."""
    return dir_name


@pytest.mark.parametrize("dir_name", listdir(instruct_dir), ids=generate_id)
def test_instruction_cases(dir_name):
    """Check that the instructions found in dir_name pass or raise the correct error.

    Parameters
    ----------
    dir_name: str
        - The directory name containing the cmd_instructions we want to test
        - Will be testing:
            - missing both .yaml and .json instructions
            - missing .json instructions
            - missing .yaml instructions
            - invalid .json instructions
            - invalid .yaml instructions
            - newer .yaml file than the .json file
            - newer .json file than the .yaml file
    """
    cmd_dir = f"{instruct_dir}/{dir_name}"
    if "invalid" in dir_name:
        # Ensure that a key error is raised
        with pytest.raises(KeyError):
            GeoipsCLI(instructions_dir=cmd_dir)
    elif "missing" in dir_name:
        if dir_name == "json_missing":
            get_instructions(ancillary_dirname=cmd_dir)
            assert exists(f"{cmd_dir}/cmd_instructions.json")
            remove(f"{cmd_dir}/cmd_instructions.json")
        else:
            with pytest.raises(FileNotFoundError):
                get_instructions(ancillary_dirname=cmd_dir)



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Unit tests on all the readers."""
import numpy as np
import pytest
from geoips.commandline.log_setup import setup_logging
from geoips.interfaces import readers
from os import environ

LOG = setup_logging()


class TestReaders:
    """Unit tests every reader."""

    available_readers = [reader_type.name for reader_type in readers.get_plugins()]

    def verify_plugin(self, plugin):
        """Yield test xarray and parameters."""
        test_xr = plugin.module.get_test_files(environ["GEOIPS_TESTDATA_DIR"])
        test_param = plugin.module.get_test_parameters()
        self.verify_xarray(test_xr, test_param)

    def verify_xarray(self, inxr, test_parameters):
        """Test every parameter."""
        for test_parameter in test_parameters:
            data_key = test_parameter["data_key"]
            data_var = test_parameter["data_var"]
            if "mean" in test_parameter:
                mean = test_parameter["mean"]
                assert np.isclose(inxr[data_key].variables[data_var].mean(), mean)

            assert inxr
            assert inxr[data_key]
            assert inxr[data_key].longitude.max()
            assert inxr[data_key].longitude.min()
            assert inxr[data_key].latitude.max()
            assert inxr[data_key].latitude.min()
            assert inxr[data_key].variables[data_var].mean()
            assert inxr["METADATA"].attrs["source_name"]
            assert inxr["METADATA"].attrs["platform_name"]
            assert inxr["METADATA"].attrs["data_provider"]
            assert inxr["METADATA"].attrs["start_datetime"]
            assert inxr["METADATA"].attrs["end_datetime"]
            assert inxr["METADATA"].attrs["interpolation_radius_of_influence"]

    @pytest.mark.parametrize("reader_name", available_readers)
    def test_reader_plugins(self, reader_name):
        """Unit test every plugin, xfail for ones with no unit tests."""
        reader = readers.get_plugin(reader_name)
        if not hasattr(reader.module, "get_test_files") or not hasattr(
            reader.module, "get_test_parameters"
        ):
            pytest.xfail(reader_name + " has no test modules")
        try:
            self.verify_plugin(reader)
        except FileNotFoundError:
            pytest.xfail(reader_name + "is missing test data")
        except ValueError as e:
            if "Input files inconsistent." in str(e):
                pytest.xfail(reader_name + "is missing test data")
            else:
                raise e



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Standard GeoIPS xarray dictionary based FCI NetCDF data reader."""

from datetime import datetime
import logging
import numpy as np
import xarray

from geoips.interfaces import readers
from geoips.plugins.modules.readers.utils.geostationary_geolocation import (
    get_geolocation,
)
from geoips.utils.context_managers import import_optional_dependencies

with import_optional_dependencies(loglevel="info"):
    """Attempt to import a package and print to LOG.info if the import fails."""
    from satpy.scene import Scene

    # Need to uncomment if HDF5_PLUGIN_PATH env variable is not set upstream,
    # though you will still not be able to interact with the data after using Scene.load
    # Until this works out of the box with out the need to set HDF5_PLUGIN_PATH
    # import hdf5plugin

LOG = logging.getLogger(__name__)

interface = "readers"
family = "standard"
name = "fci_netcdf"
source_names = ["fci"]

# Needed until the satpy devs can fix this issue:
# https://github.com/pytroll/satpy/issues/3067
# Used to convert from wavenumber (mW·m⁻²·sr⁻¹·(cm⁻¹)⁻¹)
# to wavelength (W·m⁻²·sr⁻¹·µm⁻¹)
RADIANCE_CONVERSION_COEFFICIENTS = {
    "B01Rad": 50.26570892,
    "B02Rad": 38.74919891,
    "B03Rad": 24.66601944,
    "B04Rad": 13.51755047,
    "B05Rad": 11.98462009,
    "B06Rad": 5.261388779,
    "B07Rad": 3.84947896,
    "B08Rad": 1.965775013,
    "B09Rad": 0.697183013,
    "B10Rad": 0.2536683083,
    "B11Rad": 0.1817249954,
    "B12Rad": 0.1306722015,
    "B13Rad": 0.1071347967,
    "B14Rad": 0.0899727419,
    "B15Rad": 0.06602230668,
    "B16Rad": 0.05665054172,
    "HRB03Rad": 24.66601944,
    "HRB09Rad": 0.697183013,
    "HRB14Rad": 0.0899727419,
}

DATASET_INFO = {
    # HIGH resolution = 0.5km at nadir
    "HIGH": ["vis_06_hr", "nir_22_hr"],
    # MED resolution = 1.0km at nadir
    "MED": [
        "vis_04",
        "vis_05",
        "vis_06",
        "vis_08",
        "vis_09",
        "nir_13",
        "nir_16",
        "nir_22",
        "ir_38_hr",
        "ir_105_hr",
    ],
    # LOW resolution = 2.0km at nadir
    "LOW": ["ir_38", "wv_63", "wv_73", "ir_87", "ir_97", "ir_105", "ir_123", "ir_133"],
}

BAND_MAP = {
    "B01Ref": "vis_04",
    "B01Rad": "vis_04",
    "B02Ref": "vis_05",
    "B02Rad": "vis_05",
    "B03Ref": "vis_06",
    "B03Rad": "vis_06",
    "B04Ref": "vis_08",
    "B04Rad": "vis_08",
    "B05Ref": "vis_09",
    "B05Rad": "vis_09",
    "B06Ref": "nir_13",
    "B06Rad": "nir_13",
    "B07Ref": "nir_16",
    "B07Rad": "nir_16",
    "B08Ref": "nir_22",
    "B08Rad": "nir_22",
    "B09BT": "ir_38",
    "B09Rad": "ir_38",
    "B10BT": "wv_63",
    "B10Rad": "wv_63",
    "B11BT": "wv_73",
    "B11Rad": "wv_73",
    "B12BT": "ir_87",
    "B12Rad": "ir_87",
    "B13BT": "ir_97",
    "B13Rad": "ir_97",
    "B14BT": "ir_105",
    "B14Rad": "ir_105",
    "B15BT": "ir_123",
    "B15Rad": "ir_123",
    "B16BT": "ir_133",
    "B16Rad": "ir_133",
    "HRB03Ref": "vis_06_hr",
    "HRB03Rad": "vis_06_hr",
    "HRB09BT": "ir_38_hr",
    "HRB09Rad": "ir_38_hr",
    "HRB14BT": "ir_105_hr",
    "HRB14Rad": "ir_105_hr",
}

CHAN_MAP = {val: key for key, val in BAND_MAP.items()}

ALL_GEO_VARS = [
    "solar_zenith_angle",
    "satellite_zenith_angle",
    "solar_azimuth_angle",
    "satellite_azimuth_angle",
    "latitude",
    "longitude",
]


def get_geolocation_metadata(xarray_attrs):
    """Get geolocation metadata."""
    geomet = {}
    geomet["platform_name"] = xarray_attrs["platform_name"]
    geomet["source_name"] = xarray_attrs["sensor"]  # sensor == fci
    # Used for cached filenames
    geomet["scan_mode"] = xarray_attrs["area"].area_id
    geomet["scene"] = geomet["scan_mode"]
    geomet["lon0"] = xarray_attrs["orbital_parameters"]["satellite_nominal_longitude"]
    geomet["num_lines"] = xarray_attrs["area"].y_size
    geomet["num_samples"] = xarray_attrs["area"].x_size
    geomet["start_datetime"] = xarray_attrs["start_time"]
    geomet["end_datetime"] = xarray_attrs["end_time"]
    geomet["H_km"] = 42164
    geomet["H_m"] = geomet["H_km"] * 1000.0  # Satellite altitude (m)
    geomet["BADVALS"] = {"Off_Of_Disk": -999.9}
    return geomet


def get_standard_geoips_attrs(file_attrs, area_def):
    """Set the standard geoips attrs.

    Parameters
    ----------
    file_attrs : dict
        Dictionary holding attributes found in source file

    Returns
    -------
    dict
        Dictionary of all source file attrs, along with required GeoIPS attrs
    """
    # Native metadata in file
    all_attrs = file_attrs.copy()
    all_attrs["platform_name"] = file_attrs["platform"].replace("MTI", "MTG-I")
    all_attrs["source_name"] = file_attrs["data_source"].lower()
    all_attrs["data_provider"] = "EUMETSAT"
    date_time_position = datetime.strptime(
        all_attrs["date_time_position"], "%Y%m%d%H%M%S"
    )
    # Time coverage start/end will differ across each individual granule that
    # comprises a full scan. So instead of setting the start/end datetime to the
    # time coverage start/end attr in the file metadata, use the date_time_position
    # attr, which is consistent across all files.
    # all_attrs["start_datetime"] = datetime.strptime(
    #     file_attrs["time_coverage_start"], "%Y%m%d%H%M%S"
    # )
    # all_attrs["end_datetime"] = datetime.strptime(
    #     file_attrs["time_coverage_end"], "%Y%m%d%H%M%S"
    # )
    all_attrs["start_datetime"] = date_time_position
    all_attrs["end_datetime"] = date_time_position
    all_attrs["interpolation_radius_of_influence"] = 50 * 1000
    all_attrs["area_definition"] = area_def
    if area_def:
        area_id = area_def.area_id
    else:
        area_id = None
    all_attrs["area_id"] = area_id
    return all_attrs


def load_metadata_single_file(fci_file, group=None, variable=None):
    """Load metadata from an FCI file.

    Parameters
    ----------
    fci_file : str
        Full file path to file
    group : str, optional
        Name of group to load attributes, by default None
    variable : _type_, optional
        Name of variable to load attributes, by default None

    Returns
    -------
    dict
        All metadata attrs for group/variable
    """
    all_meta_attrs = {}
    if not group:
        group = ""
    load_group = []
    for grp in group.split("/"):
        meta = xarray.open_dataset(fci_file, group="/".join(load_group)).attrs
        for key, val in meta.items():
            if key not in all_meta_attrs:
                all_meta_attrs[key] = val
        load_group.append(grp)
    if variable and variable in meta:
        for key, val in meta[variable].attrs:
            if key not in all_meta_attrs:
                all_meta_attrs[key] = val
    return all_meta_attrs


def avail_chans_in_file(fci_file):
    """Return list of available channels/variables in file.

    Parameters
    ----------
    fci_file : str
        Full file path to FCI file

    Returns
    -------
    list
        list of strings of available variables
    """
    meta = xarray.open_dataset(fci_file)
    return list(meta.l1c_channels_present.data)


def final_calibrated_variable_name(ncvar_name, chan_map=CHAN_MAP):
    """Determine the final name of dataset variable that should be used in GeoIPS.

    Parameters
    ----------
    ncvar_name : str
        Name of variable in source file
    chan_map : dict, optional
        Mapped band name used by GeoIPS, by default CHAN_MAP

    Returns
    -------
    str
        Variable name used in GeoIPS (e.g. B14BT)
    """
    if any([x in ncvar_name for x in ["vis_", "nir_"]]):
        cal_name = "Ref"
    else:
        cal_name = "BT"
    return chan_map[ncvar_name] + cal_name


def scene_to_xarray(fci_files, geoips_chans, metadata=None, area_def=None):
    """Load files into satpy.Scene object then convert to xarray.Dataset.

    Parameters
    ----------
    fci_files : list
        List of FCI netCDF files on disk
    geoips_chans : list
        Channels/variables denoted in GeoIPS nomenclature to load into memory, where
        'GeoIPS nomenclature' is the common variable mapping used in our geostationary
        readers. See 'BAND_MAP' variable for an example of these channels. Allows us
        to read in radiances, reflectances, or brightness temperatures even though the
        variable in the dataset itself (such as 'vis_04') does not provide that
        information.
    metadata : dict, optional
        Standard geoips attributes, by default None
    area_def : pyresample.AreaDefinition, optional
        Crop data to specific area definition, by default None

    Returns
    -------
    xarray.Dataset
        Calibrated FCI MTG data
    """
    scn = Scene(filenames=fci_files, reader="fci_l1c_nc")
    for gchan in geoips_chans:
        rchan = BAND_MAP.get(gchan)
        if rchan is None:
            # Most likely encountered a geolocation variable which will be computed
            # later
            continue
        if "Rad" in gchan:
            calibration = "radiance"
        elif "Ref" in gchan:
            calibration = "reflectance"
        elif "BT" in gchan:
            calibration = "brightness_temperature"
        else:
            # Star denotes the highest calibration, which is BT I think.
            calibration = "*"
        scn.load([rchan], calibration=calibration)
        # We might be deriving data from the same variable multiple times
        # I.e. B01Rad and B01Ref. Rename these data queries to the GeoIPS variable name
        # so we don't have conflicts.
        scn[gchan] = scn[rchan].copy(deep=True)
        # Delete the satpy named variable
        del scn[rchan]

    # Resample all variables in this scene to the finest resolution area definition
    # found. scn.resample() defaults to scn.finest_area()
    resampled_scn = scn.resample(resampler="native")
    LOG.interactive("Resampled scene to finest resolution.")
    if area_def:
        resampled_scn = resampled_scn.crop(area=area_def)

    xr = resampled_scn.to_xarray_dataset()
    lons, lats = resampled_scn.finest_area().get_lonlats()
    # if area_def:
    #     scn = scn.crop(area=area_def)

    # xr = scn.to_xarray_dataset()
    # lons, lats = scn.finest_area().get_lonlats()
    # Mask inf values with -999.9 for compatibility with get_geolocation func
    lons[np.isinf(lons)] = -999.9
    lats[np.isinf(lats)] = -999.9
    geo_meta = get_geolocation_metadata(xr.attrs)
    gvars = get_geolocation(xr.start_time, geo_meta, lats, lons, geo_meta["BADVALS"])
    xr["latitude"] = (("y", "x"), lats)
    xr["longitude"] = (("y", "x"), lons)
    for gkey, gvar in gvars.items():
        xr[gkey] = (("y", "x"), gvar)
    # Use a set here because gvars could include the same variables as geoips_chans
    # For example, we request satellite and solar zenith angle in our geoips_chans,
    # which is already included in gvars
    for dkey in set(geoips_chans + list(gvars)):
        # Flip data arrays upside down - otherwise image will be upside down in
        # unprojected output
        LOG.debug("Flipping data for %s", dkey)
        xr[dkey].data = np.flipud(xr[dkey].compute().data)
        # Convert from wavenumber to wavelength
        if dkey.endswith("Rad"):
            xr[dkey].data = xr[dkey].data * RADIANCE_CONVERSION_COEFFICIENTS[dkey]

    # Temporarily convert reflectances from % ranging from 0-100 to 0-1
    # This is required until gamma correction can handle reflecances from 0-100
    # This will happen when all readers return reflectance ranging from 0-100
    for dkey in list(xr.variables.keys()):
        if "Ref" in dkey:
            xr[dkey] /= 100
    # Add all standard GeoIPS attrs, if not already present
    if not metadata:
        metadata = get_standard_geoips_attrs(
            load_metadata_single_file(fci_files[0], group="data"), area_def
        )
        metadata["source_file_names"] = fci_files
    for smeta, sval in metadata.items():
        if smeta not in xr.attrs:
            xr.attrs[smeta] = sval
    return xr


def satpy_read_fci_netcdf_files(
    fci_files,
    metadata_only=False,
    chans=None,
    area_def=None,
    self_register=False,
):
    """Read and calibrate data using satpy's fci_l1c_nc reader.

    Parameters
    ----------
    fci_files : list
        list of files for a given scan time
    metadata_only : bool, optional
        Only return metadata, by default False
    chans : list, optional
        List of channels/variables to load, by default None
    area_def : pyresample.AreaDefinition, optional
        Read in data for specific area definition, by default None
    self_register : bool, optional
        Self register data to given resolution, by default False
        (currently unsupported)

    Returns
    -------
    dict
        dictionary of xarray datasets
    """
    if area_def and self_register:
        raise ValueError(
            "sector_definition and self_register are mutually exclusive keywords"
        )
    avail_vars = avail_chans_in_file(fci_files[0])
    if not chans:
        chans = avail_vars
    ingested = {}
    standard_meta = get_standard_geoips_attrs(
        load_metadata_single_file(fci_files[0], group="data"), area_def
    )
    standard_meta["source_file_names"] = fci_files
    if "METADATA" not in ingested:
        ingested["METADATA"] = xarray.Dataset(attrs=standard_meta)
    if metadata_only:
        return ingested
    # Loading all chans with satpy will not maintain native resolution for each chan.
    # This is probably OK for now when self-registering (makes things easier as satpy
    # returns all data at the same size of the lowest available resolution)
    if self_register:
        adname = "FULL_DISK"
        ingested[adname] = scene_to_xarray(fci_files, chans, metadata=standard_meta)
    # Otherwise read in chans with similar resolutions to maintain resolution
    else:
        for res, rchans in DATASET_INFO.items():
            # Find the indices of variables that correspond to the current resolution
            read_idxs = np.array([BAND_MAP.get(x, x) in rchans for x in chans])
            if not any(read_idxs) and not metadata_only:
                LOG.info("No channels to read at %s resolution", res)
                continue
            # Use those indices to retrieve the correct 'geoips' variables that we'll
            # need to read
            read_geoips_chans = list(np.array(chans)[read_idxs])
            LOG.info("Reading in following chans: %s: %s", res, read_geoips_chans)
            # Ingest the returned xarray dataset to the current resolution of the
            # xarray_dictionary
            ingested[res] = scene_to_xarray(
                fci_files, read_geoips_chans, metadata=standard_meta, area_def=area_def
            )
    return ingested


def call(
    fnames,
    metadata_only=False,
    chans=None,
    area_def=None,
    self_register=False,
):
    """Read and calibrate FCI data.

    Parameters
    ----------
    fci_files : list
        list of files for a given scan time
    metadata_only : bool, optional
        Only return metadata, by default False
    chans : list, optional
        List of channels/variables to load, by default None
    area_def : pyresample area definition, optional
        Read in data for specific area definition, by default None
        (currently unsupported)
    self_register : bool, optional
        Self register data to given resolution, by default False
        (currently unsupported)

    Returns
    -------
    dict
        dictionary of xarray datasets
    """
    fci_data = readers.read_data_to_xarray_dict(
        fnames,
        satpy_read_fci_netcdf_files,
        metadata_only,
        chans,
        area_def,
        self_register,
    )

    return fci_data



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Utilities for creating a database of tropical cyclone tracks."""

from geoips.filenames.base_paths import PATHS as gpaths
import logging

import sqlite3

LOG = logging.getLogger(__name__)

TC_DECKS_DB = gpaths["TC_DECKS_DB"]
TC_DECKS_DIR = gpaths["TC_DECKS_DIR"]


def open_tc_db(dbname=TC_DECKS_DB):
    """Open the TC Decks Database, create it if it doesn't exist."""
    # Make sure the directory exists.  If the db doesn't exist,
    # the sqlite3.connect command will create it - which will
    # fail if the directory doesn't exist.
    from os.path import dirname as pathdirname
    from geoips.filenames.base_paths import make_dirs

    make_dirs(pathdirname(dbname))

    conn = sqlite3.connect(dbname)
    LOG.interactive("Opening tc db: %s", dbname)
    conn_cursor = conn.cursor()
    # Try to create the table - if it already exists, it will just fail
    # trying to create, pass, and return the already opened db.
    try:
        conn_cursor.execute(
            """CREATE TABLE tc_trackfiles
            (id INTEGER PRIMARY KEY AUTOINCREMENT NOT NULL,
                filename text,
                last_updated timestamp DEFAULT CURRENT_TIMESTAMP NOT NULL,
                storm_num integer,
                storm_basin text,
                start_datetime timestamp,
                start_lat real,
                start_lon real,
                start_vmax real,
                start_name real,
                vmax real,
                end_datetime timestamp)"""
        )
        # Add in at some point?
        # storm_start_datetime timestamp,
    except sqlite3.OperationalError:
        pass
    return conn_cursor, conn


def check_db(filenames=None, process=False):
    """Check TC database for passed filenames.

    filenames is a list of filenames and directories. if a list element is a
    string directory name, it expands to list of files in dir.
    """
    from os.path import join as pathjoin
    from glob import glob

    if filenames is None:
        filenames = glob(pathjoin(TC_DECKS_DIR, "*"))

    updated_files = []
    cc, conn = open_tc_db()

    # We might want to rearrange this so we don't open up every file...
    # Check timestamps first.
    for filename in filenames:
        try:
            updated_files += update_fields(filename, cc, conn, process=process)
        except IndexError:
            LOG.warning("Failed parsing %s", filename)
            continue

    cc.execute("SELECT * FROM tc_trackfiles")
    # data = cc.fetchall()
    conn.close()
    # return data
    LOG.interactive("%s updated storms", len(updated_files))
    return updated_files


def update_fields(tc_trackfilename, cc, conn, process=False):
    """Update fields in TC track database with passed tc_trackfilename."""
    # Must be of form similar to
    # Gal912016.dat

    import re
    from datetime import datetime, timedelta
    from os.path import basename as pathbasename
    from os import stat as osstat

    updated_files = []

    LOG.info("Checking " + tc_trackfilename + " ... process " + str(process))

    # Check if we match Gxxdddddd.dat filename format.
    # If not just return and don't do anything.
    if (
        not re.compile(r"G\D\D\d\d\d\d\d\d\.\d\d\d\d\d\d\d\d\d\d.dat").match(
            pathbasename(tc_trackfilename)
        )
        and not re.compile(r"G\D\D\d\d\d\d\d\d\.dat").match(
            pathbasename(tc_trackfilename)
        )
        and not re.compile(r"b\D\D\d\d\d\d\d\d\.\d\d\d\d\d\d\d\d\d\d.dat").match(
            pathbasename(tc_trackfilename)
        )
        and not re.compile(r"b\D\D\d\d\d\d\d\d\.dat").match(
            pathbasename(tc_trackfilename)
        )
    ):
        LOG.info("")
        LOG.warning(
            "    DID NOT MATCH REQUIRED FILENAME FORMAT, SKIPPING: %s", tc_trackfilename
        )
        return []

    # Get all fields for the database entry for the current filename
    cc.execute("SELECT * FROM tc_trackfiles WHERE filename = ?", (tc_trackfilename,))
    data = cc.fetchone()

    file_timestamp = datetime.fromtimestamp(osstat(tc_trackfilename).st_mtime)
    # Reads timestamp out as string - convert to datetime object.
    # Check if timestamp on file is newer than timestamp in database -
    # if not, just return and don't do anything.
    if data:
        try:
            database_timestamp = datetime.strptime(
                cc.execute(
                    "SELECT last_updated from tc_trackfiles WHERE filename = ?",
                    (tc_trackfilename,),
                ).fetchone()[0],
                "%Y-%m-%d %H:%M:%S.%f",
            )
        except ValueError as resp:
            LOG.exception(f"Failed on {tc_trackfilename}")
            raise (ValueError(f"FAILED ON {tc_trackfilename}: {resp}"))
        if file_timestamp < database_timestamp:
            LOG.info("")
            LOG.interactive(
                "%s already in %s and up to date, not doing anything",
                tc_trackfilename,
                TC_DECKS_DB,
            )
            return []

    lines = open(tc_trackfilename, "r").readlines()
    # Remove any whitespace from the loaded deck file lines - otherwise will fail
    # parsing datetime information downstream
    lines = [x.replace(" ", "") for x in lines]
    start_line = lines[0].split(",")
    # Start 24h prior to start in sectorfile, for initial processing
    # storm_start_datetime = datetime.strptime(start_line[2],'%Y%m%d%H')
    start_datetime = datetime.strptime(start_line[2], "%Y%m%d%H") - timedelta(hours=24)
    end_datetime = datetime.strptime(lines[-1].split(",")[2], "%Y%m%d%H")
    start_vmax = start_line[8]
    vmax = 0
    for line in lines:
        currv = line.split(",")[8]
        track = line.split(",")[4]
        if currv and track == "BEST" and float(currv) > vmax:
            vmax = float(currv)

    if data and database_timestamp < file_timestamp:
        LOG.interactive("")
        LOG.interactive(
            "Updating start/end datetime and last_updated fields for "
            + tc_trackfilename
            + " in "
            + TC_DECKS_DB
        )
        old_start_datetime, old_end_datetime, old_vmax = cc.execute(
            "SELECT start_datetime,end_datetime,vmax from tc_trackfiles WHERE "
            "filename = ?",
            (tc_trackfilename,),
        ).fetchone()
        # Eventually add in storm_start_datetime
        # old_storm_start_datetime,old_start_datetime,old_end_datetime,old_vmax =
        #   cc.execute("SELECT storm_start_datetime,start_datetime,end_datetime,vmax
        #   from tc_trackfiles WHERE filename = ?", (tc_trackfilename,)).fetchone()
        if old_start_datetime == start_datetime.strftime("%Y-%m-%d %H:%M:%S"):
            LOG.interactive("    UNCHANGED start_datetime: " + old_start_datetime)
        else:
            LOG.interactive(
                "    Old start_datetime: "
                + old_start_datetime
                + " to new: "
                + start_datetime.strftime("%Y-%m-%d %H:%M:%S")
            )
            updated_files += [tc_trackfilename]
        # if old_storm_start_datetime ==
        #                            storm_start_datetime.strftime('%Y-%m-%d %H:%M:%S'):
        #    LOG.info('    UNCHANGED storm_start_datetime: '+old_storm_start_datetime)
        # else:
        #    LOG.info('    Old storm_start_datetime: '+old_storm_start_datetime+' to
        #              new: '+storm_start_datetime.strftime('%Y-%m-%d %H:%M:%S'))
        if old_end_datetime == end_datetime.strftime("%Y-%m-%d %H:%M:%S"):
            LOG.interactive("    UNCHANGED end_datetime: " + old_end_datetime)
        else:
            LOG.interactive(
                "    Old end_datetime: "
                + old_end_datetime
                + " to new: "
                + end_datetime.strftime("%Y-%m-%d %H:%M:%S")
            )
            updated_files += [tc_trackfilename]
        if database_timestamp == file_timestamp:
            LOG.interactive(
                "    UNCHANGED last_updated: "
                + database_timestamp.strftime("%Y-%m-%d %H:%M:%S")
            )
        else:
            LOG.interactive(
                "    Old last_updated: "
                + database_timestamp.strftime("%Y-%m-%d %H:%M:%S")
                + " to new: "
                + file_timestamp.strftime("%Y-%m-%d %H:%M:%S")
            )
            updated_files += [tc_trackfilename]
        if old_vmax == vmax:
            LOG.interactive("    UNCHANGED vmax: " + str(old_vmax))
        else:
            LOG.interactive("    Old vmax: " + str(old_vmax) + " to new: " + str(vmax))
            updated_files += [tc_trackfilename]
        cc.execute(
            """UPDATE tc_trackfiles SET
                        last_updated=?,
                        start_datetime=?,
                        end_datetime=?,
                        vmax=?
                      WHERE filename = ?""",
            # Eventually add in ?
            # storm_start_datetime=?,
            (
                file_timestamp,
                # storm_start_datetime,
                start_datetime,
                end_datetime,
                str(vmax),
                tc_trackfilename,
            ),
        )
        conn.commit()
        return updated_files

    start_lat = start_line[6].replace("N", "")
    if "S" in start_lat:
        start_lat = "-" + start_lat.replace("S", "")
    start_lon = start_line[7].replace("E", "")
    if "W" in start_lon:
        start_lon = "-" + start_lon.replace("W", "")
    storm_basin = start_line[0]
    storm_num = start_line[1]
    if re.compile(r"b\D\D\d\d\d\d\d\d\.dat").match(
        pathbasename(tc_trackfilename)
    ) or re.compile(r"b\D\D\d\d\d\d\d\d\.\d\d\d\d\d\d\d\d\d\d.dat").match(
        pathbasename(tc_trackfilename)
    ):
        # For the most part, b-deck files line up fairly with G-deck files, but the
        # index of the start name is wildly different. I almost wonder if we should
        # use the b-deck parser to grab all this information, but this is hopefully
        # good enough
        start_name = start_line[27]
    else:
        # Keep the old logic for parsing the start_name out of the G-deck files
        try:
            start_name = start_line[48] + start_line[49]
        except IndexError:
            start_name = start_line[41]

    if data is None:
        # print '    Adding '+tc_trackfilename+' to '+TC_DECKS_DB
        cc.execute(
            """insert into tc_trackfiles(
                        filename,
                        last_updated,
                        vmax,
                        storm_num,
                        storm_basin,
                        start_datetime,
                        start_lat,
                        start_lon,
                        start_vmax,
                        start_name,
                        end_datetime) values(?, ?,?, ?,?,?,?,?,?,?,?)""",
            # Eventually add in ?
            # end_datetime) values(?, ?, ?,?, ?,?,?,?,?,?,?,?)''',
            # storm_start_datetime,
            (
                tc_trackfilename,
                file_timestamp,
                str(vmax),
                storm_num,
                storm_basin,
                # storm_start_datetime,
                start_datetime,
                start_lat,
                start_lon,
                start_vmax,
                start_name,
                end_datetime,
            ),
        )
        LOG.info("")
        LOG.interactive("    Adding " + tc_trackfilename + " to " + TC_DECKS_DB)
        updated_files += [tc_trackfilename]
        conn.commit()

        # This ONLY runs if it is a brand new storm file and we requested
        # processing. Not used right now -- includes errors.
        # if process:
        #     reprocess_storm(tc_trackfilename)
    return updated_files


# The function below was commented out as it included errors, and is not used by GeoIPS
# at this time. 9/27/23

# def reprocess_storm(tc_trackfilename):
#     """Reprocess storm tc_trackfilename, using info in TC tracks database."""
#     datelist = [startdt.strftime("%Y%m%d")]
#     for nn in range((enddt - startdt).days + 2):
#         datelist += [(startdt + timedelta(nn)).strftime("%Y%m%d")]

#     hourlist = []
#     for ii in range(24):
#         hourlist += [(enddt - timedelta(hours=ii)).strftime("%H")]
#     hourlist.sort()
#     # Do newest first
#     datelist.sort(reverse=True)

#     for sat, sensor in [
#         ("gcom-w1", "amsr2"),
#         ("gpm", "gmi"),
#         ("npp", "viirs"),
#         ("jpss-1", "viirs"),
#         ("aqua", "modis"),
#         ("terra", "modis"),
#         ("himawari-8", "ahi"),
#         ("goes-16", "abi"),
#         ("goes-17", "abi"),
#     ]:
#         for datestr in datelist:
#             process_overpass(
#                 sat,
#                 sensor,
#                 productlist=None,
#                 sectorlist=[startstormsect.name],
#                 sectorfiles=None,
#                 extra_dirs=None,
#                 sector_file=sector_file,
#                 datelist=[datestr],
#                 hourlist=hourlist,
#                 queue=os.getenv("DEFAULT_QUEUE"),
#                 mp_max_cpus=3,
#                 allstatic=False,
#                 alldynamic=True,
#                 # list=True will just list files and not actually run
#                 # list=True,
#                 list=False,
#                 quiet=True,
#                 start_datetime=startdt,
#                 end_datetime=enddt,
#             )


def get_all_storms_from_db(
    start_datetime,
    end_datetime,
    tc_spec_template=None,
    trackfile_parser=None,
    include_track_files=False,
):
    """Get all entries from all storms within a specific range of time from the TC DB.

    Parameters
    ----------
    start_datetime : datetime.datetime
        Start time of desired range
    end_datetime : datetime.datetime
        End time of desired range

    Returns
    -------
    list of pyresample Area Definitions
        List of pyresample Area Definitions, each storm location that falls
        within the desired time range.

    Examples
    --------
    >>> startdt = datetime.strptime('20200216', '%Y%m%d')
    >>> enddt = datetime.strptime('20200217', '%Y%m%d')
    >>> get_storm_from_db(startdt, enddt)
    """
    from os.path import exists as path_exists

    return_area_defs = []
    connection_cursor, connection = open_tc_db()
    LOG.interactive(
        "Getting all storms from tcdb from '%s' to '%s'",
        start_datetime,
        end_datetime,
    )
    LOG.info("connection: %s", connection)
    try:
        connection_cursor.execute(
            """SELECT filename from tc_trackfiles WHERE
                                     end_datetime >= ?
                                     AND start_datetime <= ?""",
            (start_datetime, end_datetime),
        )
    except sqlite3.InterfaceError as resp:
        LOG.exception(f"{resp} error getting fields from database, return empty list")
        return return_area_defs
    deck_filenames = connection_cursor.fetchall()
    from geoips.sector_utils.tc_tracks import trackfile_to_area_defs

    for deck_filename in deck_filenames:
        if deck_filename is not None:
            # Is a tuple
            (deck_filename,) = deck_filename
        LOG.info("deck_filename %s", deck_filename)
        if not path_exists(deck_filename):
            LOG.warning("Deck file does not exist! %s", deck_filename)
            continue
        area_defs = trackfile_to_area_defs(
            deck_filename,
            trackfile_parser=trackfile_parser,
            tc_spec_template=tc_spec_template,
        )
        for area_def in area_defs:
            if (
                area_def.sector_start_datetime > start_datetime
                and area_def.sector_start_datetime < end_datetime
            ):
                if include_track_files:
                    return_area_defs += [(area_def, deck_filename)]
                else:
                    return_area_defs += [area_def]
    LOG.interactive("%s storms found in time range in tcdb!", len(return_area_defs))
    # return None if no storm matched
    return return_area_defs



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Passive Microwave 37 GHz Colorized Brightness Temperature.

Data manipulation steps for the "color37" product.
This algorithm expects Brightness Temperatures in units of degrees Kelvin
"""
import logging

LOG = logging.getLogger(__name__)

interface = "algorithms"
family = "list_numpy_to_numpy"
name = "pmw_color37"


def call(arrays):
    """color37 product algorithm data manipulation steps.

    This algorithm expects Brightness Temperatures in units of degrees Kelvin,
    and returns red green and blue gun arrays.

    Parameters
    ----------
    data : list of numpy.ndarray
        * list of numpy.ndarray or numpy.MaskedArray of channel data,
            in order of channels list above
        * Degrees Kelvin

    Returns
    -------
    numpy.ndarray
        numpy.ndarray or numpy.MaskedArray of qualitative RGBA image output
    """
    h37 = arrays[0]
    v37 = arrays[1]

    from geoips.data_manipulations.corrections import apply_data_range, apply_gamma

    red = 2.181 * v37 - 1.181 * h37
    red = apply_data_range(
        red,
        260.0,
        280.0,
        min_outbounds="crop",
        max_outbounds="crop",
        norm=True,
        inverse=True,
    )
    red = apply_gamma(red, 1.0)

    grn = (v37 - 180.0) / (300.0 - 180.0)
    grn = apply_data_range(
        grn,
        0.0,
        1.0,
        min_outbounds="crop",
        max_outbounds="crop",
        norm=True,
        inverse=False,
    )
    grn = apply_gamma(grn, 1.0)

    blu = (h37 - 160.0) / (300.0 - 160.0)
    blu = apply_data_range(
        blu,
        0.0,
        1.0,
        min_outbounds="crop",
        max_outbounds="crop",
        norm=True,
        inverse=False,
    )
    blu = apply_gamma(blu, 1.0)

    from geoips.image_utils.mpl_utils import alpha_from_masked_arrays, rgba_from_arrays

    alp = alpha_from_masked_arrays([red, grn, blu])
    rgba = rgba_from_arrays(red, grn, blu, alp)
    return rgba



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Module containing wave height colormap with transitions at 0,2,4,6,8."""

import logging

LOG = logging.getLogger(__name__)

interface = "colormappers"
family = "matplotlib"
name = "wave_height_cmap"


def call(data_range=[0, 8]):
    """Generate appropriate matplotlib colors for plotting standard wave heights.

    wind_radii_transitions contains hard coded transition values for different
    colors, in order to have consistent imagery across all sensors / products.

    Parameters
    ----------
    data_range : list of float, default=[0, 200]
        * Min and max value for colormap.
        * Ensure the data range matches the range of the algorithm specified for
          use with this colormap
        * This colormap MUST include 0 and 200

    Returns
    -------
    mpl_colors_info : dict
        Dictionary of matplotlib plotting parameters, to ensure consistent
        image output
    """
    from geoips.image_utils.colormap_utils import create_linear_segmented_colormap

    min_wave_height = data_range[0]
    max_wave_height = data_range[1]
    transition_vals = [
        (min_wave_height, 2),
        (2, 4),
        (4, 6),
        (6, max_wave_height),
    ]
    transition_colors = [
        ("white", "#08306b"),
        ("#9e9ac8", "#3f007d"),
        ("#e6f3cd", "#8ACE00"),
        ("yellow", "#de9d12"),
    ]

    # ticks = [xx[0] for xx in transition_vals]
    ticks = [0, 2, 4, 6, 8]

    LOG.info("Setting cmap")
    mpl_cmap = create_linear_segmented_colormap(
        "windspeed_cmap",
        min_wave_height,
        max_wave_height,
        transition_vals,
        transition_colors,
    )

    LOG.info("Setting norm")
    from matplotlib.colors import Normalize

    mpl_norm = Normalize(vmin=min_wave_height, vmax=max_wave_height)

    cbar_label = "Wave Height (m)"

    # Must be uniform or proportional, None not valid for Python 3
    cbar_spacing = "proportional"
    mpl_tick_labels = None
    mpl_boundaries = None

    # from geoips.image_utils.mpl_utils import create_colorbar
    # only create colorbar for final imagery
    # cbar = create_colorbar(fig, mpl_cmap, mpl_norm, ticks, cbar_label=cbar_label)
    mpl_colors_info = {
        "cmap": mpl_cmap,
        "norm": mpl_norm,
        "cbar_ticks": ticks,
        "cbar_tick_labels": mpl_tick_labels,
        "cbar_label": cbar_label,
        "boundaries": mpl_boundaries,
        "cbar_spacing": cbar_spacing,
        "colorbar": True,
        "cbar_full_width": True,
    }

    # return cbar, min_wind_speed, max_wind_speed
    return mpl_colors_info
